# Copyright 2018 NTT DATA
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from tacker import context
from tacker.db.common_services import common_services_db_plugin
from tacker.objects import heal_vnf_request
from tacker.plugins.common import constants
from tacker.tests.unit import base
from tacker.vnfm.infra_drivers.openstack import vdu


vnf_dict = {
    'attributes': {
        'heat_template': {
            'outputs': {
                'mgmt_ip-VDU1': {
                    'value': {
                        'get_attr': [
                            'CP1', 'fixed_ips', 0, 'ip_address']
                    }
                }
            },
            'description': 'Demo example\n',
            'parameters': {},
            'resources': {
                'VDU1': {
                    'type': 'OS::Nova::Server',
                    'properties': {
                        'user_data_format': 'SOFTWARE_CONFIG',
                        'availability_zone': 'nova',
                        'image': 'cirros-0.5.2-x86_64-disk',
                        'config_drive': False,
                        'flavor': {'get_resource': 'VDU1_flavor'},
                        'networks': [{'port': {'get_resource': 'CP1'}}]
                    }
                },
                'CP1': {
                    'type': 'OS::Neutron::Port',
                    'properties': {
                        'port_security_enabled': False,
                        'network': 'net_mgmt'
                    }
                },
                'VDU1_flavor': {
                    'type': 'OS::Nova::Flavor',
                    'properties': {'vcpus': 1, 'disk': 1, 'ram': 512}
                }
            }
        }
    },
    'status': 'ACTIVE',
    'vnfd_id': '576acf48-b9df-491d-a57c-342de660ec78',
    'tenant_id': '13d2ca8de70d48b2a2e0dbac2c327c0b',
    'vim_id': '3f41faa7-5630-47d2-9d4a-1216953c8887',
    'instance_id': 'd1121d3c-368b-4ac2-b39d-835aa3e4ccd8',
    'placement_attr': {'vim_name': 'openstack-vim'},
    'id': 'a27fc58e-66ae-4031-bba4-efede318c60b',
    'name': 'vnf_create_1'
}


class FakeHeatClient(mock.Mock):

    class Stack(mock.Mock):
        stack_status = 'CREATE_COMPLETE'
        outputs = [{'output_value': '192.168.120.31', 'description':
            'management ip address', 'output_key': 'mgmt_ip-vdu1'}]

    def create(self, *args, **kwargs):
        return {'stack': {'id': '4a4c2d44-8a52-4895-9a75-9d1c76c3e738'}}

    def get(self, id):
        return self.Stack()

    def update(self, stack_id, **kwargs):
        return self.Stack()

    def resource_mark_unhealthy(self, stack_id, resource_name,
                                mark_unhealthy, resource_status_reason):
        return self.Stack()


class TestVDU(base.TestCase):

    def setUp(self):
        super(TestVDU, self).setUp()
        self.context = context.get_admin_context()
        self._mock_heat_client()

        mock.patch('tacker.vnfm.vim_client.VimClient.get_vim').start()
        self.additional_paramas_obj = heal_vnf_request.HealVnfAdditionalParams(
            parameter='VDU1',
            cause=["Unable to reach while monitoring resource: 'VDU1'"])
        self.heal_request_data_obj = heal_vnf_request.HealVnfRequest(
            cause='VNF monitoring fails.',
            stack_id=vnf_dict['instance_id'],
            legacy_additional_params=[self.additional_paramas_obj])
        self.heal_vdu = vdu.Vdu(self.context, vnf_dict,
                           self.heal_request_data_obj)

        mock.patch('tacker.db.common_services.common_services_db_plugin.'
                   'CommonServicesPluginDb.create_event'
                   ).start()
        self._cos_db_plugin = \
            common_services_db_plugin.CommonServicesPluginDb()
        self.addCleanup(mock.patch.stopall)

    def _mock_heat_client(self):
        self.heat_client = mock.Mock(wraps=FakeHeatClient())
        fake_heat_client = mock.Mock()
        fake_heat_client.return_value = self.heat_client
        self._mock(
            'tacker.vnfm.infra_drivers.openstack.heat_client.HeatClient',
            fake_heat_client)

    @mock.patch('tacker.vnfm.vim_client.VimClient.get_vim')
    def test_heal_vdu(self, mock_get_vim):
        mock_get_vim.return_value = mock.MagicMock()

        self.heal_vdu.heal_vdu()

        self.heat_client.update.assert_called_once_with(
            stack_id=vnf_dict['instance_id'], existing=True)

        self._cos_db_plugin.create_event.assert_called_with(
            self.context, res_id=vnf_dict['id'],
            res_type=constants.RES_TYPE_VNF, res_state=vnf_dict['status'],
            evt_type=constants.RES_EVT_HEAL, tstamp=mock.ANY,
            details=("HealVnfRequest invoked to update the stack '%s'" %
                     vnf_dict['instance_id']))

    @mock.patch('tacker.vnfm.vim_client.VimClient.get_vim')
    def test_resource_mark_unhealthy(self, mock_get_vim):
        mock_get_vim.return_value = mock.MagicMock()

        self.heal_vdu._resource_mark_unhealthy()

        self.heat_client.resource_mark_unhealthy.assert_called_once_with(
            stack_id=vnf_dict['instance_id'],
            resource_name=self.additional_paramas_obj.parameter,
            mark_unhealthy=True,
            resource_status_reason=self.additional_paramas_obj.cause)

        self._cos_db_plugin.create_event.assert_called_with(
            self.context, res_id=vnf_dict['id'],
            res_type=constants.RES_TYPE_VNF, res_state=vnf_dict['status'],
            evt_type=constants.RES_EVT_HEAL, tstamp=mock.ANY,
            details="HealVnfRequest invoked to mark resource 'VDU1' "
                    "to unhealthy.")



# Copyright 2015 Brocade Communications System, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from datetime import datetime
from unittest import mock
from unittest.mock import patch

import ddt
import iso8601
from oslo_utils import uuidutils
import yaml

from tacker._i18n import _
from tacker.common import exceptions
from tacker import context
from tacker.db.common_services import common_services_db_plugin
from tacker.db.db_sqlalchemy import models
from tacker.db.nfvo import nfvo_db
from tacker.db.nfvo import ns_db
from tacker.db.vnfm import vnfm_db
from tacker.extensions import vnfm
from tacker import objects
from tacker.objects import heal_vnf_request
from tacker.plugins.common import constants
from tacker.tests.unit.conductor import fakes
from tacker.tests.unit.db import base as db_base
from tacker.tests.unit.db import utils
from tacker.tests.unit.vnflcm import fakes as vnflcm_fakes
from tacker.vnfm import monitor
from tacker.vnfm import plugin


class FakeDriverManager(mock.Mock):
    def invoke(self, *args, **kwargs):
        if 'create' in args:
            return uuidutils.generate_uuid()

        if 'get_resource_info' in args:
            return {'resources': {'name': 'dummy_vnf',
                                  'type': 'dummy',
                                  'id': uuidutils.generate_uuid()}}


class FakeVNFMonitor(mock.Mock):
    def update_vnf_with_maintenance(self, vnf_dict, maintenance_vdus):
        url = 'http://local:9890/v1.0/vnfs/%s/maintenance/%s' % (
            vnf_dict['id'], vnf_dict['tenant_id'])
        return {'url': url,
                'vdus': {'ALL': 'ad7ebc56',
                         'VDU1': '538745a0'}}


class FakeGreenPool(mock.Mock):
    pass


class FakeVimClient(mock.Mock):
    pass


class FakePlugin(mock.Mock):
    pass


class FakeException(Exception):
    pass


class TestVNFMPluginMonitor(db_base.SqlTestCase):
    def setUp(self):
        super(TestVNFMPluginMonitor, self).setUp()
        self._mock_vnf_manager()

    def _mock_vnf_manager(self):
        self._vnf_manager = mock.Mock(wraps=FakeDriverManager())
        self._vnf_manager.__contains__ = mock.Mock(
            return_value=True)
        fake_vnf_manager = mock.Mock()
        fake_vnf_manager.return_value = self._vnf_manager
        self._mock(
            'tacker.common.driver_manager.DriverManager', fake_vnf_manager)

    @mock.patch('tacker.db.vnfm.vnfm_db.VNFMPluginDb.get_vnfs')
    @mock.patch('tacker.vnfm.monitor.VNFMonitor.__run__')
    def test_init_monitoring(self, mock_run, mock_get_vnfs):
        vnf_id = uuidutils.generate_uuid()
        vnfs = [{
            'id': vnf_id,
            'vnf': {
                'id': vnf_id,
                'status': 'ACTIVE',
                'name': 'fake_vnf',
                'attributes': {
                    'monitoring_policy':
                        '{"vdus": '
                        '{"VDU1": {"ping": {"actions": {"failure": "respawn"},'
                        '"name": "ping", "parameters": {"count": 3,'
                        '"interval": 1, "monitoring_delay": 45, "timeout": 2},'
                        '"monitoring_params": {"count": 3, "interval": 1,'
                        '"monitoring_delay": 45, "timeout": 2}}}}}'}
            },
            'name': 'fake_vnf',
            'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437',
            'description': 'fake_vnf_description',
            'instance_id': 'da85ea1a-4ec4-4201-bbb2-8d9249eca7ec',
            'vnfd_id': 'eb094833-995e-49f0-a047-dfb56aaf7c4e',
            'vim_id': '6261579e-d6f3-49ad-8bc3-a9cb974778ff',
            'placement_attr': {'region': 'RegionOne'},
            'status': 'ACTIVE',
            'attributes': {
                    'monitoring_policy':
                        '{"vdus": '
                        '{"VDU1": {"ping": {"actions": {"failure": "respawn"},'
                        '"name": "ping", "parameters": {"count": 3,'
                        '"interval": 1, "monitoring_delay": 45, "timeout": 2},'
                        '"monitoring_params": {"count": 3, "interval": 1,'
                        '"monitoring_delay": 45, "timeout": 2}}}}}'},
            'mgmt_ip_address': '{"VDU1": "a.b.c.d"}',
            'deleted_at': datetime.min,
            'mgmt_ip_addresses': 'a.b.c.d'
        }]

        mock_get_vnfs.return_value = vnfs
        # NOTE(bhagyashris): VNFMonitor class is using a singleton pattern
        # and '_hosting_vnfs' is defined as a class level attribute.
        # If one of the unit test adds a VNF to monitor it will show up here
        # provided both the unit tests runs in the same process.
        # Hence, you must reinitialize '_hosting_vnfs' to empty dict.
        monitor.VNFMonitor._hosting_vnfs = dict()
        vnfm_plugin = plugin.VNFMPlugin()
        hosting_vnfs = vnfm_plugin._vnf_monitor._hosting_vnfs.values()
        hosting_vnf = list(hosting_vnfs)[0]['vnf']
        self.assertEqual('{"VDU1": "a.b.c.d"}', hosting_vnf['mgmt_ip_address'])
        self.assertEqual(1, len(hosting_vnfs))


@ddt.ddt
class TestVNFMPlugin(db_base.SqlTestCase):
    def setUp(self):
        super(TestVNFMPlugin, self).setUp()
        self.addCleanup(mock.patch.stopall)
        self.context = context.get_admin_context()
        self._mock_vim_client()
        self._stub_get_vim()
        self._mock_vnf_monitor()
        self._mock_vnf_alarm_monitor()
        self._mock_vnf_reservation_monitor()
        self._mock_vnf_maintenance_monitor()
        self._mock_vnf_maintenance_plugin()
        self._insert_dummy_vim()
        self.vnfm_plugin = plugin.VNFMPlugin()
        mock.patch('tacker.db.common_services.common_services_db_plugin.'
                   'CommonServicesPluginDb.create_event'
                   ).start()
        mock.patch('tacker.db.vnfm.vnfm_db.VNFMPluginDb._mgmt_driver_name',
                   return_value='noop').start()
        self.create = mock.patch('tacker.vnfm.infra_drivers.openstack.'
                                 'openstack.OpenStack.create',
                        return_value=uuidutils.generate_uuid()).start()
        self.create_wait = mock.patch('tacker.vnfm.infra_drivers.openstack.'
                                  'openstack.OpenStack.create_wait').start()
        self.update = mock.patch('tacker.vnfm.infra_drivers.openstack.'
                                 'openstack.OpenStack.update').start()
        self.update_wait = mock.patch('tacker.vnfm.infra_drivers.openstack.'
                                    'openstack.OpenStack.update_wait').start()
        self.delete = mock.patch('tacker.vnfm.infra_drivers.openstack.'
                                 'openstack.OpenStack.delete').start()
        self.delete_wait = mock.patch('tacker.vnfm.infra_drivers.openstack.'
                                      'openstack.OpenStack.'
                                      'delete_wait').start()
        self.scale = mock.patch('tacker.vnfm.infra_drivers.openstack.'
                                'openstack.OpenStack.scale',
                                return_value=uuidutils.generate_uuid()).start()
        self.scale_wait = mock.patch('tacker.vnfm.infra_drivers.openstack.'
                                'openstack.OpenStack.scale_wait',
                                return_value=uuidutils.generate_uuid()).start()

        self.heal_wait = mock.patch('tacker.vnfm.infra_drivers.openstack.'
                                    'openstack.OpenStack.heal_wait').start()

        def _fake_spawn(func, *args, **kwargs):
            func(*args, **kwargs)

        mock.patch.object(self.vnfm_plugin, 'spawn_n',
                          _fake_spawn).start()
        self._cos_db_plugin =\
            common_services_db_plugin.CommonServicesPluginDb()

    def _mock_vim_client(self):
        self.vim_client = mock.Mock(wraps=FakeVimClient())
        fake_vim_client = mock.Mock()
        fake_vim_client.return_value = self.vim_client
        self._mock(
            'tacker.vnfm.vim_client.VimClient', fake_vim_client)

    def _stub_get_vim(self):
        vim_obj = {'vim_id': '6261579e-d6f3-49ad-8bc3-a9cb974778ff',
                   'vim_name': 'fake_vim', 'vim_auth':
                   {'auth_url': 'http://localhost/identity', 'password':
                       'test_pw', 'username': 'test_user', 'project_name':
                       'test_project'}, 'vim_type': 'openstack'}
        self.vim_client.get_vim.return_value = vim_obj

    def _mock_vnf_monitor(self):
        self._vnf_monitor = mock.Mock(wraps=FakeVNFMonitor())
        fake_vnf_monitor = mock.Mock()
        fake_vnf_monitor.return_value = self._vnf_monitor
        self._mock(
            'tacker.vnfm.monitor.VNFMonitor', fake_vnf_monitor)

    def _mock_vnf_alarm_monitor(self):
        self._vnf_alarm_monitor = mock.Mock(wraps=FakeVNFMonitor())
        fake_vnf_alarm_monitor = mock.Mock()
        fake_vnf_alarm_monitor.return_value = self._vnf_alarm_monitor
        self._mock(
            'tacker.vnfm.monitor.VNFAlarmMonitor', fake_vnf_alarm_monitor)

    def _mock_vnf_reservation_monitor(self):
        self._vnf_reservation_mon = mock.Mock(wraps=FakeVNFMonitor())
        fake_vnf_reservation_monitor = mock.Mock()
        fake_vnf_reservation_monitor.return_value = self._vnf_reservation_mon
        self._mock(
            'tacker.vnfm.monitor.VNFReservationAlarmMonitor',
            fake_vnf_reservation_monitor)

    def _mock_vnf_maintenance_monitor(self):
        self._vnf_maintenance_mon = mock.Mock(wraps=FakeVNFMonitor())
        fake_vnf_maintenance_monitor = mock.Mock()
        fake_vnf_maintenance_monitor.return_value = self._vnf_maintenance_mon
        self._mock(
            'tacker.vnfm.monitor.VNFMaintenanceAlarmMonitor',
            fake_vnf_maintenance_monitor)

    def _mock_vnf_maintenance_plugin(self):
        self._vnf_maintenance_plugin = mock.Mock(wraps=FakePlugin())
        fake_vnf_maintenance_plugin = mock.Mock()
        fake_vnf_maintenance_plugin.return_value = self._vnf_maintenance_plugin
        self._mock(
            'tacker.plugins.fenix.FenixPlugin',
            fake_vnf_maintenance_plugin)

    def _insert_dummy_vnf_template(self):
        session = self.context.session
        vnf_template = vnfm_db.VNFD(
            id='eb094833-995e-49f0-a047-dfb56aaf7c4e',
            tenant_id='ad7ebc56538745a08ef7c5e97f8bd437',
            name='fake_template',
            description='fake_template_description',
            template_source='onboarded',
            deleted_at=datetime.min)
        session.add(vnf_template)
        session.flush()
        return vnf_template

    def _insert_dummy_vnf_template_inline(self):
        session = self.context.session
        vnf_template = vnfm_db.VNFD(
            id='d58bcc4e-d0cf-11e6-bf26-cec0c932ce01',
            tenant_id='ad7ebc56538745a08ef7c5e97f8bd437',
            name='tmpl-koeak4tqgoqo8cr4-dummy_inline_vnf',
            description='inline_fake_template_description',
            deleted_at=datetime.min,
            template_source='inline')
        session.add(vnf_template)
        session.flush()
        return vnf_template

    def _insert_dummy_vnf(self, status="ACTIVE"):
        session = self.context.session
        vnf_db = vnfm_db.VNF(
            id='6261579e-d6f3-49ad-8bc3-a9cb974778fe',
            tenant_id='ad7ebc56538745a08ef7c5e97f8bd437',
            name='fake_vnf',
            description='fake_vnf_description',
            instance_id='da85ea1a-4ec4-4201-bbb2-8d9249eca7ec',
            vnfd_id='eb094833-995e-49f0-a047-dfb56aaf7c4e',
            vim_id='6261579e-d6f3-49ad-8bc3-a9cb974778ff',
            placement_attr={'region': 'RegionOne'},
            status=status,
            deleted_at=datetime.min)
        session.add(vnf_db)
        session.flush()
        return vnf_db

    def _insert_dummy_pending_vnf(self, context, status='PENDING_DELETE'):
        session = context.session
        vnf_db = vnfm_db.VNF(
            id='6261579e-d6f3-49ad-8bc3-a9cb974778fe',
            tenant_id='ad7ebc56538745a08ef7c5e97f8bd437',
            name='fake_vnf',
            description='fake_vnf_description',
            instance_id='da85ea1a-4ec4-4201-bbb2-8d9249eca7ec',
            vnfd_id='eb094833-995e-49f0-a047-dfb56aaf7c4e',
            vim_id='6261579e-d6f3-49ad-8bc3-a9cb974778ff',
            placement_attr={'region': 'RegionOne'},
            status=status,
            deleted_at=datetime.min)
        session.add(vnf_db)
        session.flush()
        return vnf_db

    def _insert_scaling_attributes_vnf(self):
        session = self.context.session
        vnf_attributes = vnfm_db.VNFAttribute(
            id='7800cb81-7ed1-4cf6-8387-746468522651',
            vnf_id='6261579e-d6f3-49ad-8bc3-a9cb974778fe',
            key='scaling_group_names',
            value='{"SP1": "G1"}'
        )
        session.add(vnf_attributes)
        session.flush()
        return vnf_attributes

    def _insert_scaling_attributes_vnfd(self, invalid_policy_type=False):
        session = self.context.session
        vnfd_attributes = vnfm_db.VNFDAttribute(
            id='7800cb81-7ed1-4cf6-8387-746468522650',
            vnfd_id='eb094833-995e-49f0-a047-dfb56aaf7c4e',
            key='vnfd',
            value=utils.vnfd_scale_tosca_template
        )
        session.add(vnfd_attributes)
        session.flush()
        if invalid_policy_type:
            vnfd_template = yaml.safe_load(vnfd_attributes.value)
            vnfd_template['topology_template']['policies'][0]['SP1']['type'] \
                = "test_invalid_policy_type"
            vnfd_attributes.value = yaml.dump(vnfd_template)
        return vnfd_attributes

    def _insert_dummy_vim(self):
        session = self.context.session
        vim_db = nfvo_db.Vim(
            id='6261579e-d6f3-49ad-8bc3-a9cb974778ff',
            tenant_id='ad7ebc56538745a08ef7c5e97f8bd437',
            name='fake_vim',
            description='fake_vim_description',
            type='test_vim',
            status='Active',
            deleted_at=datetime.min,
            placement_attr={'regions': ['RegionOne']})
        vim_auth_db = nfvo_db.VimAuth(
            vim_id='6261579e-d6f3-49ad-8bc3-a9cb974778ff',
            password='encrypted_pw',
            auth_url='http://localhost/identity',
            vim_project={'name': 'test_project'},
            auth_cred={'username': 'test_user', 'user_domain_id': 'default',
                       'project_domain_id': 'default'})
        session.add(vim_db)
        session.add(vim_auth_db)
        session.flush()

    @mock.patch('tacker.vnfm.plugin.toscautils.updateimports')
    @mock.patch('tacker.vnfm.plugin.ToscaTemplate')
    @mock.patch('tacker.vnfm.plugin.toscautils.get_mgmt_driver')
    def test_create_vnfd(self, mock_get_mgmt_driver, mock_tosca_template,
                        mock_update_imports):
        mock_get_mgmt_driver.return_value = 'noop'
        mock_tosca_template.return_value = mock.ANY

        vnfd_obj = utils.get_dummy_vnfd_obj()
        result = self.vnfm_plugin.create_vnfd(self.context, vnfd_obj)
        self.assertIsNotNone(result)
        self.assertIn('id', result)
        self.assertEqual('dummy_vnfd', result['name'])
        self.assertEqual('dummy_vnfd_description', result['description'])
        self.assertEqual('noop', result['mgmt_driver'])
        self.assertIn('service_types', result)
        self.assertIn('attributes', result)
        self.assertIn('created_at', result)
        self.assertIn('updated_at', result)
        self.assertIn('template_source', result)
        yaml_dict = yaml.safe_load(utils.tosca_vnfd_openwrt)
        mock_tosca_template.assert_called_once_with(
            a_file=False, yaml_dict_tpl=yaml_dict)
        mock_get_mgmt_driver.assert_called_once_with(mock.ANY)
        mock_update_imports.assert_called_once_with(yaml_dict)
        self._cos_db_plugin.create_event.assert_called_once_with(
            self.context, evt_type=constants.RES_EVT_CREATE, res_id=mock.ANY,
            res_state=constants.RES_EVT_ONBOARDED,
            res_type=constants.RES_TYPE_VNFD, tstamp=mock.ANY)

    def test_create_vnfd_without_tosca_definitions_version(self):
        vnfd_obj = utils.get_dummy_vnfd_obj()
        vnfd_obj['vnfd']['attributes']['vnfd'].pop('tosca_definitions_version')
        self.assertRaises(exceptions.Invalid,
                          self.vnfm_plugin.create_vnfd,
                          self.context, vnfd_obj)

    def test_create_vnfd_with_empty_description(self):
        vnfd_obj = utils.get_dummy_vnfd_obj()
        vnfd_obj['vnfd']['description'] = ''
        result = self.vnfm_plugin.create_vnfd(self.context, vnfd_obj)
        self.assertIsNotNone(result)
        # If vnfd description is an empty string, it sets the description of
        # vnfd to the description that is present in the vnfd tosca template.
        self.assertEqual(yaml.safe_load(
            vnfd_obj['vnfd']['attributes']['vnfd'])['description'],
            result['description'])

    def test_create_vnfd_empty_name(self):
        vnfd_obj = utils.get_dummy_vnfd_obj()
        vnfd_obj['vnfd']['name'] = ''
        result = self.vnfm_plugin.create_vnfd(self.context, vnfd_obj)
        self.assertIsNotNone(result)
        # If vnfd name is an empty string, it sets the name of vnfd to
        # the name that is present in the vnfd tosca template.
        self.assertEqual(yaml.safe_load(vnfd_obj['vnfd']['attributes']
                    ['vnfd'])['metadata']['template_name'], result['name'])

    def test_create_vnfd_with_tosca_parser_failure(self):
        vnfd_obj = utils.get_invalid_vnfd_obj()
        self.assertRaises(vnfm.ToscaParserFailed,
                          self.vnfm_plugin.create_vnfd,
                          self.context, vnfd_obj)

    def test_create_vnfd_no_service_types(self):
        vnfd_obj = utils.get_dummy_vnfd_obj()
        vnfd_obj['vnfd'].pop('service_types')
        self.assertRaises(vnfm.ServiceTypesNotSpecified,
                          self.vnfm_plugin.create_vnfd,
                          self.context, vnfd_obj)

    def test_create_vnfd_without_dict_type_attributes(self):
        vnfd_obj = utils.get_dummy_vnfd_obj()
        # Convert dict to string.
        vnfd_obj['vnfd']['attributes']['vnfd'] = str(
            vnfd_obj['vnfd']['attributes']['vnfd'])
        self.assertRaises(vnfm.InvalidAPIAttributeType,
                          self.vnfm_plugin.create_vnfd,
                          self.context, vnfd_obj)

    def test_create_vnf_sync(self):
        self._insert_dummy_vnf_template()
        vnf_obj = utils.get_dummy_vnf_obj()
        vnf_dict = self.vnfm_plugin.create_vnf_sync(self.context,
                                                    vnf_obj['vnf'])
        self.assertIsNotNone(vnf_dict)
        self.assertEqual('ACTIVE', vnf_dict['status'])
        self._cos_db_plugin.create_event.assert_called_with(
            self.context, evt_type=constants.RES_EVT_CREATE, res_id=mock.ANY,
            res_state=mock.ANY, res_type=constants.RES_TYPE_VNF,
            tstamp=mock.ANY, details=mock.ANY)

    def test_create_vnf_with_vnfd(self):
        self._insert_dummy_vnf_template()
        vnf_obj = utils.get_dummy_vnf_obj()
        result = self.vnfm_plugin.create_vnf(self.context, vnf_obj)
        self.assertIsNotNone(result)
        self.assertIn('id', result)
        self.assertIn('instance_id', result)
        self.assertIn('status', result)
        self.assertIn('attributes', result)
        self.assertIn('mgmt_ip_address', result)
        self.assertIn('created_at', result)
        self.assertIn('updated_at', result)
        self.assertEqual('ACTIVE', result['status'])
        self._cos_db_plugin.create_event.assert_called_with(
            self.context, evt_type=constants.RES_EVT_CREATE, res_id=mock.ANY,
            res_state=mock.ANY, res_type=constants.RES_TYPE_VNF,
            tstamp=mock.ANY, details=mock.ANY)

    @mock.patch('tacker.vnfm.plugin.VNFMPlugin.create_vnfd')
    def test_create_vnf_from_template(self, mock_create_vnfd):
        self._insert_dummy_vnf_template_inline()
        mock_create_vnfd.return_value = {'id':
                'd58bcc4e-d0cf-11e6-bf26-cec0c932ce01'}
        vnf_obj = utils.get_dummy_inline_vnf_obj()
        result = self.vnfm_plugin.create_vnf(self.context, vnf_obj)
        self.assertIsNotNone(result)
        self.assertIn('id', result)
        self.assertIn('instance_id', result)
        self.assertIn('status', result)
        self.assertIn('attributes', result)
        self.assertIn('mgmt_ip_address', result)
        self.assertIn('created_at', result)
        self.assertIn('updated_at', result)
        self.assertEqual('ACTIVE', result['status'])
        mock_create_vnfd.assert_called_once_with(mock.ANY, mock.ANY)
        self._cos_db_plugin.create_event.assert_called_with(
            self.context, evt_type=constants.RES_EVT_CREATE,
            res_id=mock.ANY,
            res_state=mock.ANY, res_type=constants.RES_TYPE_VNF,
            tstamp=mock.ANY, details=mock.ANY)

    def test_create_vnf_with_param_values(self):
        self._insert_dummy_vnf_template()
        vnf_obj = utils.get_dummy_vnf_obj()
        vnf_obj['vnf']['attributes'] = {'param_values':
        {'image_name': 'cirros-0.5.2-x86_64-disk', 'flavor': 'm1.tiny'}}
        result = self.vnfm_plugin.create_vnf(self.context, vnf_obj)
        self.assertIsNotNone(result)
        self.assertEqual(vnf_obj['vnf']['attributes']['param_values'],
                         result['attributes']['param_values'])
        self.assertEqual('ACTIVE', result['status'])

    def test_create_vnf_with_config_option(self):
        self._insert_dummy_vnf_template()
        vnf_obj = utils.get_dummy_vnf_obj()
        config = utils.get_dummy_vnf_config_obj()
        vnf_obj['vnf']['attributes'] = config['vnf']['attributes']
        result = self.vnfm_plugin.create_vnf(self.context, vnf_obj)
        self.assertEqual(vnf_obj['vnf']['attributes']['config'],
                         result['attributes']['config'])
        self.assertEqual('ACTIVE', result['status'])

    def test_create_vnf_fail_with_invalid_infra_driver_exception(self):
        self.vim_client.get_vim.return_value['vim_type'] = 'test_invalid_vim'
        self._insert_dummy_vnf_template()
        vnf_obj = utils.get_dummy_vnf_obj()
        self.assertRaises(vnfm.InvalidInfraDriver,
                          self.vnfm_plugin.create_vnf,
                          self.context, vnf_obj)

    def test_create_vnf_with_invalid_param_and_config_format(self):
        self._insert_dummy_vnf_template()
        vnf_obj = utils.get_dummy_vnf_obj()
        vnf_obj['vnf']['attributes']['param_values'] = 'image_name'
        vnf_obj['vnf']['attributes']['config'] = 'test'
        self.assertRaises(vnfm.InvalidAPIAttributeType,
                          self.vnfm_plugin.create_vnf,
                          self.context, vnf_obj)

    @patch('tacker.vnfm.plugin.VNFMPlugin._delete_vnf')
    def test_create_vnf_fail(self, mock_delete_vnf):
        self._insert_dummy_vnf_template()
        vnf_obj = utils.get_dummy_vnf_obj()
        self.create.side_effect = vnfm.HeatClientException(msg='test')
        self.assertRaises(vnfm.HeatClientException,
                          self.vnfm_plugin.create_vnf,
                          self.context, vnf_obj)
        vnf_id = self.vnfm_plugin._delete_vnf.call_args[0][1]
        mock_delete_vnf.assert_called_once_with(self.context, vnf_id,
                                                force_delete=True)

    def test_create_vnf_create_wait_failed_exception(self):
        self._insert_dummy_vnf_template()
        vnf_obj = utils.get_dummy_vnf_obj()
        self.create_wait.side_effect = vnfm.VNFCreateWaitFailed(
            reason="failed")
        vnf_dict = self.vnfm_plugin.create_vnf(self.context, vnf_obj)
        self.assertEqual(constants.ERROR,
                         vnf_dict['status'])

    @patch('tacker.vnfm.plugin.VNFMPlugin.mgmt_call')
    def test_create_vnf_mgmt_driver_exception(self, mock_mgmt_call):
        self._insert_dummy_vnf_template()
        vnf_obj = utils.get_dummy_vnf_obj()
        mock_mgmt_call.side_effect = exceptions.MgmtDriverException
        vnf_dict = self.vnfm_plugin.create_vnf(self.context, vnf_obj)
        self.assertEqual(constants.ERROR,
                         vnf_dict['status'])

    @patch('tacker.db.vnfm.vnfm_db.VNFMPluginDb._create_vnf_post')
    @patch('tacker.db.vnfm.vnfm_db.VNFMPluginDb._create_vnf_pre')
    @patch('tacker.db.vnfm.vnfm_db.VNFMPluginDb._create_vnf_status')
    def test_create_vnf_with_alarm_url(self, mock_create_vnf_status,
                                       mock_create_vnf_pre,
                                       mock_create_vnf_post):
        self._insert_dummy_vnf_template()
        vnf_obj = utils.get_dummy_vnf_obj()
        alarm_url_dict = {'vdu_hcpu_usage_scaling_out':
                        'http://localhost/identity',
                          'vdu_lcpu_usage_scaling_in':
                        'http://localhost/identity'}
        self._vnf_alarm_monitor.update_vnf_with_alarm.return_value = \
            alarm_url_dict
        dummy_vnf = self._get_dummy_vnf(utils.vnfd_alarm_scale_tosca_template,
                                        status='PENDING_CREATE')
        mock_create_vnf_pre.return_value = dummy_vnf
        vnf_dict = self.vnfm_plugin.create_vnf(self.context, vnf_obj)
        self.assertEqual(alarm_url_dict['vdu_lcpu_usage_scaling_in'],
                         vnf_dict['attributes']['vdu_lcpu_usage_scaling_in'])
        self.assertEqual(alarm_url_dict['vdu_hcpu_usage_scaling_out'],
                         vnf_dict['attributes']['vdu_hcpu_usage_scaling_out'])

    @patch('tacker.vnfm.plugin.VNFMPlugin._create_vnf_wait')
    def test_show_vnf_details_vnf_inactive(self, mock_create_vnf_wait):
        self._insert_dummy_vnf_template()
        vnf_obj = utils.get_dummy_vnf_obj()
        result = self.vnfm_plugin.create_vnf(self.context, vnf_obj)
        self.assertRaises(vnfm.VNFInactive, self.vnfm_plugin.get_vnf_resources,
                          self.context, result['id'])

    @patch('tacker.vnfm.infra_drivers.openstack.openstack.OpenStack.'
           'get_resource_info')
    def test_show_vnf_details_vnf_active(self, mock_get_resource_info):
        self._insert_dummy_vnf_template()
        active_vnf = self._insert_dummy_vnf()
        mock_get_resource_info.return_value = {'resources': {'name':
                                                            'dummy_vnf',
                                                            'type': 'dummy',
                                                            'id':
                                                uuidutils.generate_uuid()}}
        resources = self.vnfm_plugin.get_vnf_resources(self.context,
                                                       active_vnf['id'])[0]
        self.assertIn('name', resources)
        self.assertIn('type', resources)
        self.assertIn('id', resources)

    def test_delete_vnf(self):
        self._insert_dummy_vnf_template()
        dummy_vnf_obj = self._insert_dummy_vnf()
        self.vnfm_plugin.delete_vnf(self.context, dummy_vnf_obj[
            'id'])
        self._vnf_monitor.delete_hosting_vnf.assert_called_with(mock.ANY)
        self.delete.assert_called_with(plugin=mock.ANY, context=mock.ANY,
                                       vnf_id=mock.ANY,
                                       auth_attr=mock.ANY,
                                       region_name=mock.ANY)
        self._cos_db_plugin.create_event.assert_called_with(
            self.context, evt_type=constants.RES_EVT_DELETE, res_id=mock.ANY,
            res_state=mock.ANY, res_type=constants.RES_TYPE_VNF,
            tstamp=mock.ANY, details=mock.ANY)

    def test_delete_vnf_in_pending_state(self):
        # delete_vnf will raise exception when VNF status in PENDING_*
        self._insert_dummy_vnf_template()
        dummy_vnf_obj = self._insert_dummy_pending_vnf(self.context)
        self.assertRaises(vnfm.VNFDeleteFailed,
                          self.vnfm_plugin.delete_vnf,
                          self.context,
                          dummy_vnf_obj['id'])

    @ddt.data('PENDING_DELETE', 'PENDING_CREATE')
    def test_force_delete_vnf(self, status):
        self._insert_dummy_vnf_template()
        dummy_vnf_obj = self._insert_dummy_pending_vnf(self.context, status)
        vnfattr = {'vnf': {'attributes': {'force': True}}}
        self.vnfm_plugin.delete_vnf(self.context, dummy_vnf_obj[
            'id'], vnf=vnfattr)
        self._vnf_monitor.delete_hosting_vnf.assert_called_with(mock.ANY)
        self.delete.assert_called_with(plugin=mock.ANY, context=mock.ANY,
                                       vnf_id=mock.ANY,
                                       auth_attr=mock.ANY,
                                       region_name=mock.ANY)
        self._cos_db_plugin.create_event.assert_called_with(
            self.context, evt_type=constants.RES_EVT_DELETE, res_id=mock.ANY,
            res_state=mock.ANY, res_type=constants.RES_TYPE_VNF,
            tstamp=mock.ANY, details="VNF Delete Complete")

    def test_force_delete_vnf_non_admin(self):
        self._insert_dummy_vnf_template()
        non_admin_context = context.Context(user_id=None,
                                            tenant_id=None,
                                            is_admin=False)
        dummy_vnf_obj = self._insert_dummy_pending_vnf(non_admin_context)
        vnfattr = {'vnf': {'attributes': {'force': True}}}
        self.assertRaises(exceptions.AdminRequired,
                          self.vnfm_plugin.delete_vnf,
                          non_admin_context,
                          dummy_vnf_obj['id'], vnf=vnfattr)

    @patch('tacker.db.vnfm.vnfm_db.VNFMPluginDb._delete_vnf_post')
    def test_delete_vnf_fail(self, mock_delete_vnf_post):
        self.delete.side_effect = FakeException
        self._insert_dummy_vnf_template()
        dummy_device_obj = self._insert_dummy_vnf()
        self.assertRaises(FakeException,
                          self.vnfm_plugin.delete_vnf, self.context,
                          dummy_device_obj['id'])
        self._vnf_monitor.delete_hosting_vnf.assert_called_once_with(
            dummy_device_obj['id'])
        mock_delete_vnf_post.assert_called_once_with(self.context, mock.ANY,
                                                     mock.ANY)
        self._cos_db_plugin.create_event.assert_called_with(
            self.context, evt_type=constants.RES_EVT_DELETE, res_id=mock.ANY,
            res_state=mock.ANY, res_type=constants.RES_TYPE_VNF,
            tstamp=mock.ANY, details=mock.ANY)

    @patch('tacker.db.vnfm.vnfm_db.VNFMPluginDb.set_vnf_error_status_reason')
    def test_delete_vnf_delete_wait_failed_exception(self,
                                            mock_set_vnf_error_status_reason):
        self._insert_dummy_vnf_template()
        dummy_vnf_obj = self._insert_dummy_vnf()
        self.delete_wait.side_effect = vnfm.VNFDeleteWaitFailed(
            reason='failed')
        self.vnfm_plugin.delete_vnf(self.context, dummy_vnf_obj['id'])
        mock_set_vnf_error_status_reason.assert_called_once_with(self.context,
                                                                 mock.ANY,
                                                                 mock.ANY)

    def test_delete_vnf_failed_with_status_pending_create(self):
        self._insert_dummy_vnf_template()
        dummy_device_obj_with_pending_create_status = self. \
            _insert_dummy_vnf(status="PENDING_CREATE")
        self.assertRaises(vnfm.VNFInUse, self.vnfm_plugin.delete_vnf,
                          self.context,
                          dummy_device_obj_with_pending_create_status['id'])

    def _insert_dummy_ns_template(self):
        session = self.context.session
        attributes = {
            'nsd': 'imports: [VNF1, VNF2]\ntopology_template:\n  inputs:\n  '
                   '  vl1_name: {default: net_mgmt, description: name of VL1'
                   ' virtuallink, type: string}\n    vl2_name: {default: '
                   'net0, description: name of VL2 virtuallink, type: string'
                   '}\n  node_templates:\n    VL1:\n      properties:\n     '
                   '   network_name: {get_input: vl1_name}\n        vendor: '
                   'tacker\n      type: tosca.nodes.nfv.VL\n    VL2:\n      '
                   'properties:\n        network_name: {get_input: vl2_name}'
                   '\n        vendor: tacker\n      type: tosca.nodes.nfv.VL'
                   '\n    VNF1:\n      requirements:\n      - {virtualLink1: '
                   'VL1}\n      - {virtualLink2: VL2}\n      type: tosca.node'
                   's.nfv.VNF1\n    VNF2: {type: tosca.nodes.nfv.VNF2}\ntosca'
                   '_definitions_version: tosca_simple_profile_for_nfv_1_0_0'
                   '\n'}
        nsd_template = ns_db.NSD(
            id='eb094833-995e-49f0-a047-dfb56aaf7c4e',
            tenant_id='ad7ebc56538745a08ef7c5e97f8bd437',
            name='fake_template',
            vnfds={'tosca.nodes.nfv.VNF1': 'vnf1',
                   'tosca.nodes.nfv.VNF2': 'vnf2'},
            description='fake_nsd_template_description',
            deleted_at=datetime.min,
            template_source='onboarded')
        session.add(nsd_template)
        for (key, value) in attributes.items():
            attribute_db = ns_db.NSDAttribute(
                id=uuidutils.generate_uuid(),
                nsd_id='eb094833-995e-49f0-a047-dfb56aaf7c4e',
                key=key,
                value=value)
            session.add(attribute_db)
        session.flush()
        return nsd_template

    def _insert_dummy_ns(self):
        session = self.context.session
        ns = ns_db.NS(
            id='ba6bf017-f6f7-45f1-a280-57b073bf78ea',
            name='dummy_ns',
            tenant_id='ad7ebc56538745a08ef7c5e97f8bd437',
            status='ACTIVE',
            nsd_id='eb094833-995e-49f0-a047-dfb56aaf7c4e',
            vim_id='6261579e-d6f3-49ad-8bc3-a9cb974778ff',
            description='dummy_ns_description',
            vnf_ids='[5761579e-d6f3-49ad-8bc3-a9cb73477846,'
                    '6261579e-d6f3-49ad-8bc3-a9cb974778fe]',
            deleted_at=datetime.min)
        session.add(ns)
        session.flush()
        return ns

    def test_delete_vnf_of_active_ns(self):
        self._insert_dummy_ns_template()
        self._insert_dummy_ns()
        self.assertRaises(vnfm.VNFInUse, self.vnfm_plugin.delete_vnf,
            self.context, '6261579e-d6f3-49ad-8bc3-a9cb974778fe')

    def test_update_vnf(self):
        self._insert_dummy_vnf_template()
        dummy_vnf_obj = self._insert_dummy_vnf()
        vnf_config_obj = utils.get_dummy_vnf_config_obj()
        result = self.vnfm_plugin.update_vnf(self.context, dummy_vnf_obj[
            'id'], vnf_config_obj)
        self.assertIsNotNone(result)
        self.assertEqual(dummy_vnf_obj['id'], result['id'])
        self.assertIn('instance_id', result)
        self.assertIn('status', result)
        self.assertIn('attributes', result)
        self.assertIn('mgmt_ip_address', result)
        self.assertIn('updated_at', result)
        self._cos_db_plugin.create_event.assert_called_with(
            self.context, evt_type=constants.RES_EVT_UPDATE, res_id=mock.ANY,
            res_state=mock.ANY, res_type=constants.RES_TYPE_VNF,
            tstamp=mock.ANY)

    @patch('tacker.db.vnfm.vnfm_db.VNFMPluginDb._update_vnf_post')
    def test_update_vnf_with_exception(self, mock_update_vnf_post):
        self.update.side_effect = FakeException
        self._insert_dummy_vnf_template()
        dummy_device_obj = self._insert_dummy_vnf()
        vnf_config_obj = utils.get_dummy_vnf_config_obj()
        self.assertRaises(FakeException,
                          self.vnfm_plugin.update_vnf, self.context,
                          dummy_device_obj['id'], vnf_config_obj)
        self._vnf_monitor.delete_hosting_vnf.assert_called_once_with(
            dummy_device_obj['id'])
        mock_update_vnf_post.assert_called_once_with(self.context,
                                                     dummy_device_obj['id'],
                                                     constants.ERROR,
                                                     mock.ANY,
                                                     constants.PENDING_UPDATE,
                                                     constants.RES_EVT_UPDATE)

        self._cos_db_plugin.create_event.assert_called_with(
            self.context, evt_type=constants.RES_EVT_UPDATE, res_id=mock.ANY,
            res_state=mock.ANY, res_type=constants.RES_TYPE_VNF,
            tstamp=mock.ANY)

    def test_update_vnf_invalid_config_format(self):
        self._insert_dummy_vnf_template()
        dummy_vnf_obj = self._insert_dummy_vnf()
        vnf_config_obj = utils.get_dummy_vnf_config_obj()
        vnf_config_obj['vnf']['attributes']['config'] = {'vdus': {
            'vdu1': {'config': {'firewall': 'dummy_firewall_values'}}}}
        result = self.vnfm_plugin.update_vnf(self.context, dummy_vnf_obj[
            'id'], vnf_config_obj)
        self.assertEqual(constants.ACTIVE, result['status'])

    @patch('tacker.db.vnfm.vnfm_db.VNFMPluginDb.set_vnf_error_status_reason')
    @mock.patch('tacker.vnfm.plugin.VNFMPlugin.mgmt_call')
    def test_update_vnf_fail_mgmt_driver_error(self, mock_mgmt_call,
                                            mock_set_vnf_error_status_reason):
        self._insert_dummy_vnf_template()
        dummy_vnf_obj = self._insert_dummy_vnf()
        vnf_config_obj = utils.get_dummy_vnf_config_obj()
        mock_mgmt_call.side_effect = exceptions.MgmtDriverException
        vnf_dict = self.vnfm_plugin.update_vnf(self.context,
                                               dummy_vnf_obj['id'],
                                               vnf_config_obj)
        self.assertEqual(constants.ERROR,
                         vnf_dict['status'])
        mock_set_vnf_error_status_reason.assert_called_once_with(self.context,
                                                          dummy_vnf_obj['id'],
                                                   'VNF configuration failed')

    @patch('tacker.db.vnfm.vnfm_db.VNFMPluginDb.set_vnf_error_status_reason')
    def test_update_vnf_fail_update_wait_error(self,
                                            mock_set_vnf_error_status_reason):
        self._insert_dummy_vnf_template()
        dummy_vnf_obj = self._insert_dummy_vnf()
        vnf_config_obj = utils.get_dummy_vnf_config_obj()
        self.update_wait.side_effect = vnfm.VNFUpdateWaitFailed(
            reason='failed')
        self.assertRaises(vnfm.VNFUpdateWaitFailed,
                          self.vnfm_plugin.update_vnf, self.context,
                          dummy_vnf_obj['id'], vnf_config_obj)
        self._vnf_monitor.\
            delete_hosting_vnf.assert_called_once_with(dummy_vnf_obj['id'])
        mock_set_vnf_error_status_reason.assert_called_once_with(self.context,
                                                        dummy_vnf_obj['id'],
                                                    'VNF Update failed')

    def test_update_vnf_param(self):
        self._insert_dummy_vnf_template()
        dummy_device_obj = self._insert_dummy_vnf()
        vnf_param_obj = utils.get_dummy_vnf_param_obj()
        result = self.vnfm_plugin.update_vnf(self.context,
                                             dummy_device_obj['id'],
                                             vnf_param_obj)
        self.assertIsNotNone(result)
        self.assertEqual(dummy_device_obj['id'], result['id'])
        self.assertIn('instance_id', result)
        self.assertIn('status', result)
        self.assertIn('attributes', result)
        self.assertIn('mgmt_ip_address', result)
        self.assertIn('updated_at', result)
        self._cos_db_plugin.create_event.assert_called_with(
            self.context, evt_type=constants.RES_EVT_UPDATE, res_id=mock.ANY,
            res_state=mock.ANY, res_type=constants.RES_TYPE_VNF,
            tstamp=mock.ANY)

    def test_update_vnf_invalid_config_type(self):
        self._insert_dummy_vnf_template()
        dummy_device_obj = self._insert_dummy_vnf()
        vnf_param_obj = utils.get_dummy_vnf_invalid_config_type_obj()
        self.assertRaises(vnfm.InvalidAPIAttributeType,
                          self.vnfm_plugin.update_vnf,
                          self.context,
                          dummy_device_obj['id'],
                          vnf_param_obj)

    def test_update_vnf_invalid_param_type(self):
        self._insert_dummy_vnf_template()
        dummy_device_obj = self._insert_dummy_vnf()
        vnf_param_obj = utils.get_dummy_vnf_invalid_param_type_obj()
        self.assertRaises(vnfm.InvalidAPIAttributeType,
                          self.vnfm_plugin.update_vnf,
                          self.context,
                          dummy_device_obj['id'],
                          vnf_param_obj)

    def test_update_vnf_invalid_param_content(self):
        self.update.side_effect = vnfm.VNFUpdateInvalidInput(
            reason='failed')
        self._insert_dummy_vnf_template()
        dummy_device_obj = self._insert_dummy_vnf()
        vnf_param_obj = utils.get_dummy_vnf_invalid_param_content()
        self.assertRaises(vnfm.VNFUpdateInvalidInput,
                          self.vnfm_plugin.update_vnf,
                          self.context,
                          dummy_device_obj['id'],
                          vnf_param_obj)

    def _get_dummy_scaling_policy(self, type):
        vnf_scale = {}
        vnf_scale['scale'] = {}
        vnf_scale['scale']['type'] = type
        vnf_scale['scale']['policy'] = 'SP1'
        return vnf_scale

    def _get_scaling_vnf(self, type, invalid_policy_type=False):
        # create vnfd
        self._insert_dummy_vnf_template()
        self._insert_scaling_attributes_vnfd(invalid_policy_type)

        # create vnf
        dummy_vnf_obj = self._insert_dummy_vnf()
        self._insert_scaling_attributes_vnf()

        # scale vnf
        vnf_scale = self._get_dummy_scaling_policy(type)
        return dummy_vnf_obj, vnf_scale

    def _test_scale_vnf(self, type):
        dummy_vnf_obj, vnf_scale = self._get_scaling_vnf(type)
        self.vnfm_plugin.create_vnf_scale(
            self.context,
            dummy_vnf_obj['id'],
            vnf_scale)
        # validate
        self.scale.assert_called_once_with(
            plugin=mock.ANY,
            context=mock.ANY,
            auth_attr=mock.ANY,
            policy=mock.ANY,
            region_name=mock.ANY
        )
        self.scale_wait.assert_called_once_with(plugin=self.vnfm_plugin,
                                                context=self.context,
                                                auth_attr=mock.ANY,
                                                policy=mock.ANY,
                                                region_name=mock.ANY,
                                                last_event_id=mock.ANY)
        self._cos_db_plugin.create_event.assert_called_with(
            self.context,
            evt_type=constants.RES_EVT_SCALE,
            res_id='6261579e-d6f3-49ad-8bc3-a9cb974778fe',
            res_state='ACTIVE',
            res_type=constants.RES_TYPE_VNF,
            tstamp=mock.ANY)

    def test_scale_vnf_out(self):
        self._test_scale_vnf('out')

    def test_scale_vnf_in(self):
        self._test_scale_vnf('in')

    @patch('tacker.db.vnfm.vnfm_db.VNFMPluginDb._update_vnf_scaling_status')
    @patch('tacker.db.vnfm.vnfm_db.VNFMPluginDb.set_vnf_error_status_reason')
    def test_scale_vnf_with_vnf_policy_action_exception(self,
                                            mock_set_vnf_error_status_reason,
                                            mock_update_vnf_scaling_status):
        dummy_vnf_obj, vnf_scale = self._get_scaling_vnf('in')
        self.scale.side_effect = FakeException
        self.assertRaises(FakeException,
                          self.vnfm_plugin.create_vnf_scale,
                          self.context, dummy_vnf_obj['id'],
                          vnf_scale)
        mock_update_vnf_scaling_status.assert_called_with(
            self.context, mock.ANY, [constants.PENDING_SCALE_IN],
            constants.ERROR, mock.ANY)
        mock_set_vnf_error_status_reason.assert_called_with(
            self.context, dummy_vnf_obj['id'], mock.ANY)

    @mock.patch('tacker.vnfm.plugin.VNFMPlugin.get_vnf_policies')
    def test_scale_vnf_with_policy_not_found_exception(self,
                                                      mock_get_vnf_policies):
        dummy_vnf_obj, vnf_scale = self._get_scaling_vnf('in')
        mock_get_vnf_policies.return_value = None
        self.assertRaises(exceptions.VnfPolicyNotFound,
                          self.vnfm_plugin.create_vnf_scale,
                          self.context, dummy_vnf_obj['id'],
                          vnf_scale)

    def test_scale_vnf_with_invalid_policy_type(self):
        dummy_vnf_obj, vnf_scale = self._get_scaling_vnf('in',
                                                invalid_policy_type=True)
        self.assertRaises(exceptions.VnfPolicyTypeInvalid,
                          self.vnfm_plugin.create_vnf_scale,
                          self.context, dummy_vnf_obj['id'], vnf_scale)

    def test_scale_vnf_with_invalid_policy_action(self):
        dummy_vnf_obj, vnf_scale = \
            self._get_scaling_vnf('test_invalid_policy_action')
        self.assertRaises(exceptions.VnfPolicyActionInvalid,
                          self.vnfm_plugin.create_vnf_scale,
                          self.context, dummy_vnf_obj['id'], vnf_scale)

    def test_scale_vnf_scale_wait_failed_exception(self):
        dummy_vnf_obj, vnf_scale = \
            self._get_scaling_vnf('in')
        self.scale_wait.side_effect = vnfm.VNFScaleWaitFailed(
            reason='test')
        self.assertRaises(vnfm.VNFScaleWaitFailed,
                          self.vnfm_plugin.create_vnf_scale,
                          self.context, dummy_vnf_obj['id'], vnf_scale)

    def _get_dummy_vnf(self, vnfd_template, status='ACTIVE'):
        dummy_vnf = utils.get_dummy_vnf()
        dummy_vnf['vnfd']['attributes']['vnfd'] = vnfd_template
        dummy_vnf['status'] = status
        dummy_vnf['instance_id'] = '4c00108e-c69d-4624-842d-389c77311c1d'
        dummy_vnf['vim_id'] = '437ac8ef-a8fb-4b6e-8d8a-a5e86a376e8b'
        return dummy_vnf

    def _create_vnf_trigger_data(self, policy_name, action_value):
        vnf_id = "6261579e-d6f3-49ad-8bc3-a9cb974778fe"
        trigger_request = {"trigger": {"action_name": action_value, "params": {
            "credential": "026kll6n", "data": {"current": "alarm",
                                               'alarm_id':
                                    "b7fa9ffd-0a4f-4165-954b-5a8d0672a35f"}},
            "policy_name": policy_name}}
        expected_result = {"action_name": action_value, "params": {
            "credential": "026kll6n", "data": {"current": "alarm",
            "alarm_id": "b7fa9ffd-0a4f-4165-954b-5a8d0672a35f"}},
            "policy_name": policy_name}
        return vnf_id, trigger_request, expected_result

    @patch('tacker.vnfm.policy_actions.autoscaling.autoscaling.'
           'VNFActionAutoscaling.execute_action')
    def _test_create_vnf_trigger(self, mock_execute_action,
                                 policy_name, action_value):
        vnf_id, trigger_request, expected_result = self.\
            _create_vnf_trigger_data(policy_name, action_value)
        self._vnf_alarm_monitor.process_alarm_for_vnf.return_value = True
        trigger_result = self.vnfm_plugin.create_vnf_trigger(self.context,
                                                    vnf_id, trigger_request)
        self.assertEqual(expected_result, trigger_result)

    @patch('tacker.db.vnfm.vnfm_db.VNFMPluginDb.get_vnf')
    def test_create_vnf_trigger_respawn(self, mock_get_vnf):
        dummy_vnf = self._get_dummy_vnf(
            utils.vnfd_alarm_respawn_tosca_template)
        mock_get_vnf.return_value = dummy_vnf
        self._test_create_vnf_trigger(policy_name="vdu_hcpu_usage_respawning",
                                      action_value="respawn")

    @patch('tacker.db.vnfm.vnfm_db.VNFMPluginDb.get_vnf')
    def test_create_vnf_trigger_scale(self, mock_get_vnf):
        dummy_vnf = self._get_dummy_vnf(
            utils.vnfd_alarm_scale_tosca_template)
        mock_get_vnf.return_value = dummy_vnf
        self._test_create_vnf_trigger(policy_name="vdu_hcpu_usage_scaling_out",
                                      action_value="SP1-out")

    @patch('tacker.db.vnfm.vnfm_db.VNFMPluginDb.get_vnf')
    def test_create_vnf_trigger_multi_actions(self, mock_get_vnf):
        dummy_vnf = self._get_dummy_vnf(
            utils.vnfd_alarm_multi_actions_tosca_template)
        mock_get_vnf.return_value = dummy_vnf
        self._test_create_vnf_trigger(policy_name="mon_policy_multi_actions",
                                      action_value="respawn&log")

    @patch('tacker.db.vnfm.vnfm_db.VNFMPluginDb.get_vnf')
    def test_create_vnf_trigger_without_policy_actions(self, mock_get_vnf):
        dummy_vnf = self._get_dummy_vnf(
            utils.vnfd_alarm_multi_actions_tosca_template)
        mock_get_vnf.return_value = dummy_vnf
        vnf_id, trigger_request, _ = self._create_vnf_trigger_data(
            "mon_policy_multi_actions", "respawn&log")
        self._vnf_alarm_monitor.process_alarm_for_vnf.return_value = False
        self.assertRaises(exceptions.AlarmUrlInvalid,
                          self.vnfm_plugin.create_vnf_trigger,
                          self.context,
                          vnf_id, trigger_request)

    @patch('tacker.db.vnfm.vnfm_db.VNFMPluginDb.get_vnf')
    def test_create_vnf_trigger_with_invalid_policy_name(self, mock_get_vnf):
        dummy_vnf = self._get_dummy_vnf(
            utils.vnfd_alarm_multi_actions_tosca_template)
        mock_get_vnf.return_value = dummy_vnf
        vnf_id, trigger_request, _ = self._create_vnf_trigger_data(
            "invalid_policy_name", "respawn&log")
        self.assertRaises(exceptions.TriggerNotFound,
                          self.vnfm_plugin.create_vnf_trigger,
                          self.context,
                          vnf_id, trigger_request)

    @patch('tacker.db.vnfm.vnfm_db.VNFMPluginDb.get_vnf')
    @patch('tacker.vnfm.plugin.LOG')
    def test_create_vnf_trigger_scale_with_invalid_vnf_status(self,
                                                    mock_log, mock_get_vnf):
        dummy_vnf = self._get_dummy_vnf(utils.vnfd_alarm_scale_tosca_template)
        dummy_vnf['status'] = "PENDING_CREATE"
        mock_get_vnf.return_value = dummy_vnf
        vnf_id, trigger_request, expected_result = self. \
            _create_vnf_trigger_data("vdu_hcpu_usage_scaling_out", "SP1-out")
        expected_error_msg = (_("Scaling Policy action skipped due to status"
                                ' %(status)s for vnf %(vnfid)s') %
                              {'status': dummy_vnf['status'],
                               'vnfid': dummy_vnf['id']})
        self._vnf_alarm_monitor.process_alarm_for_vnf.return_value = True
        trigger_result = self.vnfm_plugin.create_vnf_trigger(self.context,
                                                     vnf_id, trigger_request)

        mock_log.info.assert_called_with(expected_error_msg)
        self.assertEqual(expected_result, trigger_result)

    @patch('tacker.db.vnfm.vnfm_db.VNFMPluginDb.get_vnf')
    def test_get_vnf_policies(self, mock_get_vnf):
        vnf_id = "6261579e-d6f3-49ad-8bc3-a9cb974778fe"
        dummy_vnf = self._get_dummy_vnf(
            utils.vnfd_alarm_respawn_tosca_template)
        mock_get_vnf.return_value = dummy_vnf
        policies = self.vnfm_plugin.get_vnf_policies(self.context, vnf_id,
            filters={'name': 'vdu1_cpu_usage_monitoring_policy'})
        self.assertEqual(1, len(policies))

    @mock.patch('tacker.vnfm.plugin.toscautils.get_mgmt_driver')
    def test_mgmt_driver(self, mock_get_mgmt_driver):
        mock_get_mgmt_driver.return_value = 'dummy_mgmt_driver'

        vnfd_obj = utils.get_dummy_vnfd_obj()
        self.assertRaises(vnfm.InvalidMgmtDriver,
                          self.vnfm_plugin.create_vnfd,
                          self.context, vnfd_obj)

    @mock.patch('tacker.vnfm.plugin.VNFMPlugin.get_vnf_policies')
    def test_get_vnf_policy_by_type(self, mock_get_vnf_policies):
        mock_get_vnf_policies.return_value = None

        self.assertRaises(exceptions.VnfPolicyTypeInvalid,
                          self.vnfm_plugin.get_vnf_policy_by_type,
                          self.context,
                          uuidutils.generate_uuid(),
                          policy_type='invalid_policy_type')

    @patch('tacker.vnfm.infra_drivers.openstack.openstack.OpenStack.'
           'heal_vdu')
    @mock.patch('tacker.vnfm.monitor.VNFMonitor.update_hosting_vnf')
    def test_heal_vnf_vdu(self, mock_update_hosting_vnf, mock_heal_vdu):
        self._insert_dummy_vnf_template()
        dummy_device_obj = self._insert_dummy_vnf()
        additional_params_obj = heal_vnf_request.HealVnfAdditionalParams(
            parameter='VDU1',
            cause=["Unable to reach while monitoring resource: 'VDU1'"])
        heal_request_data_obj = heal_vnf_request.HealVnfRequest(
            stack_id=dummy_device_obj['instance_id'],
            cause='VNF monitoring fails.',
            legacy_additional_params=[additional_params_obj])
        result = self.vnfm_plugin.heal_vnf(self.context,
                                           dummy_device_obj['id'],
                                           heal_request_data_obj)
        self.assertIsNotNone(result)
        self.assertEqual(dummy_device_obj['id'], result['id'])
        self.assertIn('instance_id', result)
        self.assertIn('status', result)
        self.assertIn('attributes', result)
        self.assertIn('mgmt_ip_address', result)
        self.assertIn('updated_at', result)
        self.assertEqual('ACTIVE', result['status'])
        mock_heal_vdu.assert_called_with(plugin=self.vnfm_plugin,
            context=self.context, vnf_dict=mock.ANY,
            heal_request_data_obj=heal_request_data_obj)

    @patch('tacker.db.vnfm.vnfm_db.VNFMPluginDb.get_vnf')
    def test_create_vnf_trigger_scale_with_reservation(self, mock_get_vnf):
        dummy_vnf = self._get_dummy_vnf(
            utils.vnfd_instance_reservation_alarm_scale_tosca_template)
        mock_get_vnf.return_value = dummy_vnf
        self._test_create_vnf_trigger(policy_name="start_actions",
                                      action_value="SP_RSV-out")

    def test_create_placement_constraint(self):
        res_str = '[{"id_type": "RES_MGMT", "resource_id": ' + \
            '"2c6e5cc7-240d-4458-a683-1fe648351200", ' + \
            '"vim_connection_id": ' + \
            '"2a63bee3-0c43-4568-bcfa-b0cb733e064c"}]'
        placemnt = models.PlacementConstraint(
            id='c2947d8a-2c67-4e8f-ad6f-c0889b351c17',
            vnf_instance_id='7ddc38c3-a116-48b0-bfc1-68d7f306f467',
            affinity_or_anti_affinity='ANTI_AFFINITY',
            scope='ZONE',
            server_group_name='my_compute_placement_policy',
            resource=res_str,
            deleted_at=datetime.min)
        pls_list = [placemnt]
        vnf_inst = models.VnfInstance(
            id='7ddc38c3-a116-48b0-bfc1-68d7f306f467',
            vnf_provider=' ',
            vnf_product_name=' ',
            vnf_software_version=' ',
            vnfd_version=' ',
            vnfd_id='8d86480e-d4e6-4ee0-ba4d-08217118d6cb',
            instantiation_state=' ',
            tenant_id='9b3f0518-bf6b-4982-af32-d282ce577c8f',
            created_at=datetime(
                2020, 1, 1, 1, 1, 1,
                tzinfo=iso8601.UTC),
            vnf_pkg_id=uuidutils.generate_uuid())
        self.context.session.add(vnf_inst)
        self.context.session.flush()

        self.vnfm_plugin.create_placement_constraint(
            self.context, pls_list)

    def test_get_placement_constraint(self):
        res_str = '[{"id_type": "RES_MGMT", "resource_id": ' + \
            '"2c6e5cc7-240d-4458-a683-1fe648351200", ' + \
            '"vim_connection_id": ' + \
            '"2a63bee3-0c43-4568-bcfa-b0cb733e064c"}]'
        placemnt = models.PlacementConstraint(
            id='c2947d8a-2c67-4e8f-ad6f-c0889b351c17',
            vnf_instance_id='7ddc38c3-a116-48b0-bfc1-68d7f306f467',
            affinity_or_anti_affinity='ANTI_AFFINITY',
            scope='ZONE',
            server_group_name='my_compute_placement_policy',
            resource=res_str,
            deleted_at=datetime.min)
        vnf_inst = models.VnfInstance(
            id='7ddc38c3-a116-48b0-bfc1-68d7f306f467',
            vnf_provider=' ',
            vnf_product_name=' ',
            vnf_software_version=' ',
            vnfd_version=' ',
            vnfd_id='8d86480e-d4e6-4ee0-ba4d-08217118d6cb',
            instantiation_state=' ',
            tenant_id='9b3f0518-bf6b-4982-af32-d282ce577c8f',
            created_at=datetime(
                2020, 1, 1, 1, 1, 1,
                tzinfo=iso8601.UTC),
            vnf_pkg_id=uuidutils.generate_uuid())
        self.context.session.add(vnf_inst)
        self.context.session.flush()
        self.context.session.add(placemnt)
        self.context.session.flush()

        res = self.vnfm_plugin.get_placement_constraint(
            self.context, '7ddc38c3-a116-48b0-bfc1-68d7f306f467')
        self.assertEqual(1, len(res))

    def test_update_placement_constraint_heal(self):
        res_str = '[{"id_type": "RES_MGMT", "resource_id": ' + \
            '"2c6e5cc7-240d-4458-a683-1fe648351200", ' + \
            '"vim_connection_id": ' + \
            '"2a63bee3-0c43-4568-bcfa-b0cb733e064c"}]'
        placemnt = models.PlacementConstraint(
            id='c2947d8a-2c67-4e8f-ad6f-c0889b351c17',
            vnf_instance_id='7ddc38c3-a116-48b0-bfc1-68d7f306f467',
            affinity_or_anti_affinity='ANTI_AFFINITY',
            scope='ZONE',
            server_group_name='my_compute_placement_policy',
            resource=res_str,
            deleted_at=datetime.min)
        res_str2 = '[{"id_type": "GRANT", "resource_id": ' + \
            '"4cef1b7e-8e5f-430e-b32e-a7585a61d61c"}]'
        placemnt2 = models.PlacementConstraint(
            id='c2947d8a-2c67-4e8f-ad6f-c0889b351c17',
            vnf_instance_id='7ddc38c3-a116-48b0-bfc1-68d7f306f467',
            affinity_or_anti_affinity='ANTI_AFFINITY',
            scope='ZONE',
            server_group_name='my_compute_placement_policy',
            resource=res_str2,
            deleted_at=datetime.min)
        vnf_inst = models.VnfInstance(
            id='7ddc38c3-a116-48b0-bfc1-68d7f306f467',
            vnf_provider=' ',
            vnf_product_name=' ',
            vnf_software_version=' ',
            vnfd_version=' ',
            vnfd_id='8d86480e-d4e6-4ee0-ba4d-08217118d6cb',
            instantiation_state=' ',
            tenant_id='9b3f0518-bf6b-4982-af32-d282ce577c8f',
            created_at=datetime(
                2020, 1, 1, 1, 1, 1,
                tzinfo=iso8601.UTC),
            vnf_pkg_id=uuidutils.generate_uuid())
        self.context.session.add(vnf_inst)
        self.context.session.flush()
        self.context.session.add(placemnt)
        self.context.session.flush()

        vnf_info = {}
        vnf_info['grant'] = objects.Grant()
        placement_obj_list = []
        placement_obj_list.append(placemnt2)
        vnf_info['placement_obj_list'] = placement_obj_list

        insta = fakes.return_vnf_instance('INSTANTIATED')
        vnfc = objects.VnfcResourceInfo()
        vnfc.id = '4cef1b7e-8e5f-430e-b32e-a7585a61d61c'
        vnfc.vdu_id = 'VDU1'
        c_rsc = objects.ResourceHandle()
        c_rsc.vim_connection_id = '2a63bee3-0c43-4568-bcfa-b0cb733e064c'
        c_rsc.resource_id = '9aa1e075-2aa1-46ce-a27a-35a581190219'
        vnfc.compute_resource = c_rsc
        insta.instantiated_vnf_info.vnfc_resource_info.append(vnfc)

        self.vnfm_plugin.update_placement_constraint_heal(
            self.context, vnf_info, insta)

    def test_delete_placement_constraint(self):
        res_str = '[{"id_type": "RES_MGMT", "resource_id": ' + \
            '"2c6e5cc7-240d-4458-a683-1fe648351200", ' + \
            '"vim_connection_id": ' + \
            '"2a63bee3-0c43-4568-bcfa-b0cb733e064c"}]'
        placemnt = models.PlacementConstraint(
            id='c2947d8a-2c67-4e8f-ad6f-c0889b351c17',
            vnf_instance_id='7ddc38c3-a116-48b0-bfc1-68d7f306f467',
            affinity_or_anti_affinity='ANTI_AFFINITY',
            scope='ZONE',
            server_group_name='my_compute_placement_policy',
            resource=res_str,
            deleted_at=datetime.min)
        vnf_inst = models.VnfInstance(
            id='7ddc38c3-a116-48b0-bfc1-68d7f306f467',
            vnf_provider=' ',
            vnf_product_name=' ',
            vnf_software_version=' ',
            vnfd_version=' ',
            vnfd_id='8d86480e-d4e6-4ee0-ba4d-08217118d6cb',
            instantiation_state=' ',
            tenant_id='9b3f0518-bf6b-4982-af32-d282ce577c8f',
            created_at=datetime(
                2020, 1, 1, 1, 1, 1,
                tzinfo=iso8601.UTC),
            vnf_pkg_id=uuidutils.generate_uuid())
        self.context.session.add(vnf_inst)
        self.context.session.flush()
        self.context.session.add(placemnt)
        self.context.session.flush()

        self.vnfm_plugin.delete_placement_constraint(
            self.context, '7ddc38c3-a116-48b0-bfc1-68d7f306f467')

    def test_update_vnf_rollback_pre_scale(self):
        vnf_info = {}
        vnf_lcm_op_occ = vnflcm_fakes.vnflcm_rollback()
        vnf_info['vnf_lcm_op_occ'] = vnf_lcm_op_occ
        vnf_info['id'] = uuidutils.generate_uuid()
        self.vnfm_plugin._update_vnf_rollback_pre(
            self.context, vnf_info)

    def test_update_vnf_rollback_pre_insta(self):
        vnf_info = {}
        vnf_lcm_op_occ = vnflcm_fakes.vnflcm_rollback_insta()
        vnf_info['vnf_lcm_op_occ'] = vnf_lcm_op_occ
        vnf_info['id'] = uuidutils.generate_uuid()
        self.vnfm_plugin._update_vnf_rollback_pre(
            self.context, vnf_info)

    def test_update_vnf_rollback_scale(self):
        vnf_info = {}
        vnf_lcm_op_occ = vnflcm_fakes.vnflcm_rollback()
        vnf_info['vnf_lcm_op_occ'] = vnf_lcm_op_occ
        vnf_info['id'] = uuidutils.generate_uuid()
        self.vnfm_plugin._update_vnf_rollback(
            self.context, vnf_info,
            'ERROR', 'ACTIVE')

    def test_update_vnf_rollback_insta(self):
        vnf_info = {}
        vnf_lcm_op_occ = vnflcm_fakes.vnflcm_rollback_insta()
        vnf_info['vnf_lcm_op_occ'] = vnf_lcm_op_occ
        vnf_info['id'] = uuidutils.generate_uuid()
        self.vnfm_plugin._update_vnf_rollback(
            self.context, vnf_info,
            'ERROR', 'INACTIVE')

    def test_update_vnf_rollback_status_err_scale(self):
        vnf_info = {}
        vnf_lcm_op_occ = vnflcm_fakes.vnflcm_rollback()
        vnf_info['vnf_lcm_op_occ'] = vnf_lcm_op_occ
        vnf_info['id'] = uuidutils.generate_uuid()
        self.vnfm_plugin.update_vnf_rollback_status_err(
            self.context, vnf_info)

    def test_update_vnf_rollback_status_err_insta(self):
        vnf_info = {}
        vnf_lcm_op_occ = vnflcm_fakes.vnflcm_rollback_insta()
        vnf_info['vnf_lcm_op_occ'] = vnf_lcm_op_occ
        vnf_info['id'] = uuidutils.generate_uuid()
        self.vnfm_plugin.update_vnf_rollback_status_err(
            self.context, vnf_info)



# Copyright 2017 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""add_attributes_to_vnffg

Revision ID: e8918cda6433
Revises: 000632983ada
Create Date: 2017-02-09 00:11:08.081746

"""

# flake8: noqa: E402

# revision identifiers, used by Alembic.
revision = 'e8918cda6433'
down_revision = '000632983ada'

from alembic import op
import sqlalchemy as sa
from tacker.db.types import Json


def upgrade(active_plugins=None, options=None):
    op.add_column('vnffgs', sa.Column('attributes', Json))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import base64
import datetime
import hashlib
import io
import json
import os
import requests
import shutil
import tempfile
import uuid
import zipfile

import ddt
from oslo_config import cfg
from requests_mock.contrib import fixture as requests_mock_fixture
from tacker import auth
from tacker.tests import base

from tacker.tests.unit.vnfm.infra_drivers.openstack.fixture_data import client
from tacker.tests.unit.vnfpkgm import fakes
from tacker.tests import utils
from tacker.tests import uuidsentinel
import tacker.vnfm.nfvo_client as nfvo_client
from unittest import mock


def _count_mock_history(history, *url):
    req_count = 0
    for mock_history in history:
        actual_url = '{}://{}'.format(mock_history.scheme,
              mock_history.hostname)
        if actual_url in url:
            req_count += 1
    return req_count


@ddt.ddt
class TestVnfPackageRequest(base.BaseTestCase):

    client_fixture_class = client.ClientFixture
    sdk_connection_fixure_class = client.SdkConnectionFixture

    def setUp(self):
        super(TestVnfPackageRequest, self).setUp()
        self.requests_mock = self.useFixture(requests_mock_fixture.Fixture())
        self.url = "http://nfvo.co.jp/vnfpkgm/v1/vnf_packages"
        self.nfvo_url = "http://nfvo.co.jp"
        self.test_package_dir = 'tacker/tests/unit/vnfm/'
        self.headers = {'Content-Type': 'application/json'}

        self.token_endpoint = 'https://oauth2/tokens'
        self.oauth_url = 'https://oauth2'
        self.auth_user_name = 'test_user'
        self.auth_password = 'test_password'

        cfg.CONF.set_override('auth_type', None,
                              group='authentication')
        cfg.CONF.set_override(
            "base_url",
            self.url,
            group='connect_vnf_packages')
        cfg.CONF.set_default(
            name='pipeline',
            group='connect_vnf_packages',
            default=[
                "package_content",
                "vnfd"])
        cfg.CONF.set_override('user_name', self.auth_user_name,
                              group='authentication')
        cfg.CONF.set_override('password', self.auth_password,
                              group='authentication')
        cfg.CONF.set_override('token_endpoint', self.token_endpoint,
                              group='authentication')
        cfg.CONF.set_override('client_id', self.auth_user_name,
                              group='authentication')
        cfg.CONF.set_override('client_password', self.auth_password,
                              group='authentication')
        auth.auth_manager = auth._AuthManager()
        nfvo_client.VnfPackageRequest._connector = nfvo_client._Connect(
            2, 1, 20)

    def tearDown(self):
        super(TestVnfPackageRequest, self).tearDown()
        self.addCleanup(mock.patch.stopall)

    def assert_auth_basic(self, acutual_request):
        actual_auth = acutual_request._request.headers.get("Authorization")
        expected_auth = base64.b64encode(
            '{}:{}'.format(
                self.auth_user_name,
                self.auth_password).encode('utf-8')).decode()
        self.assertEqual("Basic " + expected_auth, actual_auth)

    def assert_auth_client_credentials(self, acutual_request, expected_token):
        actual_auth = acutual_request._request.headers.get(
            "Authorization")
        self.assertEqual("Bearer " + expected_token, actual_auth)

    def assert_zipfile(
            self,
            actual_zip,
            expected_zips,
            expected_artifacts=None):
        expected_artifacts = expected_artifacts if expected_artifacts else []

        def check_zip(expected_zip):
            self.assertIsInstance(expected_zip, zipfile.ZipFile)
            for expected_name in expected_zip.namelist():
                expected_checksum = hashlib.sha256(
                    expected_zip.read(expected_name)).hexdigest()
                actual_checksum = hashlib.sha256(
                    actual_zip.read(expected_name)).hexdigest()
                self.assertEqual(expected_checksum, actual_checksum)

        try:
            self.assertIsInstance(actual_zip, zipfile.ZipFile)
            self.assertIsNone(actual_zip.testzip())

            expected_elm_cnt = sum(
                map(lambda x: len(x.namelist()), expected_zips))
            self.assertEqual(expected_elm_cnt +
                    len(expected_artifacts), len(actual_zip.namelist()))

            for expected_zip in expected_zips:
                check_zip(expected_zip)

            for expected_artifact in expected_artifacts:
                expected_checksum = hashlib.sha256(
                    open(expected_artifact, 'rb').read()).hexdigest()
                actual_checksum = hashlib.sha256(
                    actual_zip.read(expected_artifact)).hexdigest()
                self.assertEqual(expected_checksum, actual_checksum)
        except Exception as e:
            self.fail(e)

    def json_serial_date_to_dict(self, json_obj):
        def json_serial(obj):
            if isinstance(obj, datetime.datetime):
                return obj.isoformat()

            raise TypeError("Type %s not serializable" % type(obj))

        serial_json_str = json.dumps(json_obj, default=json_serial)
        return json.loads(serial_json_str)

    def test_init(self):
        self.assertEqual(self.url, cfg.CONF.connect_vnf_packages.base_url)
        self.assertEqual(["package_content", "vnfd"],
                         cfg.CONF.connect_vnf_packages.pipeline)
        self.assertEqual(2, cfg.CONF.connect_vnf_packages.retry_num)
        self.assertEqual(30, cfg.CONF.connect_vnf_packages.retry_wait)
        self.assertEqual(20, cfg.CONF.connect_vnf_packages.timeout)

    def _make_zip_file_from_sample(self, dir_name, read_vnfd_only=False):
        unique_name = str(uuid.uuid4())
        temp_dir = os.path.join('/tmp', unique_name)
        utils.copy_csar_files(temp_dir, dir_name, read_vnfd_only)
        tempfd, temp_filepath = tempfile.mkstemp(suffix=".zip", dir=temp_dir)
        os.close(tempfd)
        zipfile.ZipFile(temp_filepath, 'w')
        self.addCleanup(shutil.rmtree, temp_dir)
        return temp_filepath

    @ddt.data({'content': 'vnfpkgm1',
    'vnfd': None,
    'artifacts': None},
    {'content': None,
    'vnfd': 'vnfpkgm2',
    'artifacts': None},
        {'content': None,
    'vnfd': None,
    'artifacts': ["vnfd_lcm_user_data.yaml"]},
        {'content': 'vnfpkgm1',
    'vnfd': 'vnfpkgm2',
    'artifacts': ["vnfd_lcm_user_data.yaml"]},
        {'content': 'vnfpkgm1',
    'vnfd': None,
    'artifacts': None},
        {'content': None,
    'vnfd': 'vnfpkgm2',
    'artifacts': ["vnfd_lcm_user_data.yaml"]},
        {'content': 'vnfpkgm1',
    'vnfd': 'vnfpkgm2',
    'artifacts': ["vnfd_lcm_user_data.yaml"]},
    )
    @ddt.unpack
    def test_download_vnf_packages(self, content, vnfd, artifacts):
        fetch_base_url = os.path.join(self.url, uuidsentinel.vnf_pkg_id)
        expected_connect_cnt = 0
        pipelines = []

        if content:
            expected_connect_cnt += 1
            pipelines.append('package_content')
            path = self._make_zip_file_from_sample(content)
            with open(path, 'rb') as test_package_content_zip_obj:
                expected_package_content_zip = zipfile.ZipFile(
                    io.BytesIO(test_package_content_zip_obj.read()))
                test_package_content_zip_obj.seek(0)
                self.requests_mock.register_uri(
                    'GET',
                    os.path.join(
                        fetch_base_url,
                        'package_content'),
                    content=test_package_content_zip_obj.read(),
                    headers={
                        'Content-Type': 'application/zip'},
                    status_code=200)

        if vnfd:
            expected_connect_cnt += 1
            pipelines.append('vnfd')
            path = self._make_zip_file_from_sample(vnfd, read_vnfd_only=True)
            with open(path, 'rb') as test_vnfd_zip_obj:
                expected_vnfd_zip = zipfile.ZipFile(
                    io.BytesIO(test_vnfd_zip_obj.read()))
                test_vnfd_zip_obj.seek(0)
                self.requests_mock.register_uri(
                    'GET',
                    os.path.join(
                        fetch_base_url,
                        'vnfd'),
                    content=test_vnfd_zip_obj.read(),
                    headers={
                        'Content-Type': 'application/zip'},
                    status_code=200)

        if artifacts:
            pipelines.append('artifacts')
            artifacts = [os.path.join("tacker/tests/etc/samples", p)
                         for p in artifacts]
            for artifact_path in artifacts:
                expected_connect_cnt += 1
                with open(artifact_path, 'rb') as artifact_path_obj:
                    self.requests_mock.register_uri(
                        'GET',
                        os.path.join(
                            fetch_base_url,
                            'artifacts',
                            artifact_path),
                        headers={
                            'Content-Type': 'application/octet-stream'},
                        status_code=200,
                        content=artifact_path_obj.read())

        cfg.CONF.set_default(
            name='pipeline',
            group='connect_vnf_packages',
            default=pipelines)

        if artifacts:
            res = nfvo_client.VnfPackageRequest.download_vnf_packages(
                uuidsentinel.vnf_pkg_id, artifacts)
        else:
            res = nfvo_client.VnfPackageRequest.download_vnf_packages(
                uuidsentinel.vnf_pkg_id)

        history = self.requests_mock.request_history
        req_count = _count_mock_history(history, self.nfvo_url)

        self.assertIsInstance(res, io.BytesIO)
        actual_zip = zipfile.ZipFile(res)
        if content and vnfd:
            self.assert_zipfile(
                actual_zip, [
                    expected_package_content_zip,
                    expected_vnfd_zip],
                artifacts)
        elif content:
            self.assert_zipfile(
                actual_zip, [expected_package_content_zip], artifacts)
        elif vnfd:
            self.assert_zipfile(
                actual_zip, [expected_vnfd_zip], artifacts)
        else:
            self.assert_zipfile(
                actual_zip, [], artifacts)

        self.assertEqual(expected_connect_cnt, req_count)

    def test_download_vnf_packages_with_auth_basic(self):
        cfg.CONF.set_override('auth_type', 'BASIC',
                              group='authentication')
        auth.auth_manager = auth._AuthManager()

        expected_connect_cnt = \
            self._download_vnf_packages_all_pipeline_with_assert()
        history = self.requests_mock.request_history
        req_count = _count_mock_history(history, self.nfvo_url)
        self.assertEqual(expected_connect_cnt, req_count)
        for h in history:
            self.assert_auth_basic(h)

    def test_download_vnf_packages_with_auth_client_credentials(self):
        cfg.CONF.set_override('auth_type', 'OAUTH2_CLIENT_CREDENTIALS',
                              group='authentication')

        expected_connect_cnt = 1
        self.requests_mock.register_uri('POST',
            self.token_endpoint,
            json={'access_token': 'test_token', 'token_type': 'bearer'},
            headers={'Content-Type': 'application/json'},
            status_code=200)

        auth.auth_manager = auth._AuthManager()

        expected_connect_cnt += \
            self._download_vnf_packages_all_pipeline_with_assert()
        history = self.requests_mock.request_history
        req_count = _count_mock_history(history, self.nfvo_url, self.oauth_url)
        self.assertEqual(expected_connect_cnt, req_count)
        self.assert_auth_basic(history[0])
        for h in history[1:]:
            self.assert_auth_client_credentials(h, "test_token")

    def _download_vnf_packages_all_pipeline_with_assert(self):
        fetch_base_url = os.path.join(self.url, uuidsentinel.vnf_pkg_id)
        expected_connect_cnt = 0
        pipelines = []

        content = 'vnfpkgm1'
        expected_connect_cnt += 1
        pipelines.append('package_content')
        path = self._make_zip_file_from_sample(content)
        with open(path, 'rb') as test_package_content_zip_obj:
            expected_package_content_zip = zipfile.ZipFile(
                io.BytesIO(test_package_content_zip_obj.read()))
            test_package_content_zip_obj.seek(0)
            self.requests_mock.register_uri(
                'GET',
                os.path.join(
                    fetch_base_url,
                    'package_content'),
                content=test_package_content_zip_obj.read(),
                headers={
                    'Content-Type': 'application/zip'},
                status_code=200)

        vnfd = 'vnfpkgm2'
        expected_connect_cnt += 1
        pipelines.append('vnfd')
        path = self._make_zip_file_from_sample(vnfd, read_vnfd_only=True)
        with open(path, 'rb') as test_vnfd_zip_obj:
            expected_vnfd_zip = zipfile.ZipFile(
                io.BytesIO(test_vnfd_zip_obj.read()))
            test_vnfd_zip_obj.seek(0)
            self.requests_mock.register_uri(
                'GET',
                os.path.join(
                    fetch_base_url,
                    'vnfd'),
                content=test_vnfd_zip_obj.read(),
                headers={
                    'Content-Type': 'application/zip'},
                status_code=200)

        artifacts = ["vnfd_lcm_user_data.yaml"]
        pipelines.append('artifacts')
        artifacts = [os.path.join("tacker/tests/etc/samples", p)
                    for p in artifacts]
        for artifact_path in artifacts:
            expected_connect_cnt += 1
            with open(artifact_path, 'rb') as artifact_path_obj:
                self.requests_mock.register_uri(
                    'GET',
                    os.path.join(
                        fetch_base_url,
                        'artifacts',
                        artifact_path),
                    headers={
                        'Content-Type': 'application/octet-stream'},
                    status_code=200,
                    content=artifact_path_obj.read())

        cfg.CONF.set_default(
            name='pipeline',
            group='connect_vnf_packages',
            default=pipelines)

        res = nfvo_client.VnfPackageRequest.download_vnf_packages(
            uuidsentinel.vnf_pkg_id, artifacts)
        self.assertIsInstance(res, io.BytesIO)

        actual_zip = zipfile.ZipFile(res)
        self.assert_zipfile(
            actual_zip, [
                expected_package_content_zip,
                expected_vnfd_zip], artifacts)

        return expected_connect_cnt

    def test_download_vnf_packages_content_disposition(self):
        fetch_base_url = os.path.join(self.url, uuidsentinel.vnf_pkg_id)
        test_yaml_filepath = os.path.join(
            'tacker/tests/etc/samples',
            'vnfd_lcm_user_data.yaml')
        with open(test_yaml_filepath, 'rb') as test_yaml_obj:
            headers = {
                'Content-Type': 'application/octet-stream',
                'Content-Disposition':
                    'filename={}'.format(test_yaml_filepath)}
            self.requests_mock.register_uri(
                'GET',
                os.path.join(
                    fetch_base_url,
                    'vnfd'),
                content=test_yaml_obj.read(),
                headers=headers,
                status_code=200)

        cfg.CONF.set_default(
            name='pipeline',
            group='connect_vnf_packages',
            default=['vnfd'])

        res = nfvo_client.VnfPackageRequest.download_vnf_packages(
            uuidsentinel.vnf_pkg_id)
        history = self.requests_mock.request_history
        req_count = _count_mock_history(history, self.nfvo_url)
        self.assertEqual(1, req_count)

        self.assertIsInstance(res, io.BytesIO)
        actual_zip = zipfile.ZipFile(res)
        self.assert_zipfile(actual_zip, [], [test_yaml_filepath])

    def test_download_vnf_packages_non_content_disposition_raise_download(
            self):
        fetch_base_url = os.path.join(self.url, uuidsentinel.vnf_pkg_id)
        test_yaml_filepath = os.path.join(
            'tacker/tests/etc/samples',
            'vnfd_lcm_user_data.yaml')
        with open(test_yaml_filepath, 'rb') as test_yaml_obj:
            headers = {'Content-Type': 'application/octet-stream'}
            self.requests_mock.register_uri(
                'GET',
                os.path.join(
                    fetch_base_url,
                    'vnfd'),
                content=test_yaml_obj.read(),
                headers=headers,
                status_code=200)

        cfg.CONF.set_default(
            name='pipeline',
            group='connect_vnf_packages',
            default=['vnfd'])

        self.assertRaises(
            nfvo_client.FaliedDownloadContentException,
            nfvo_client.VnfPackageRequest.download_vnf_packages,
            uuidsentinel.vnf_pkg_id)

        history = self.requests_mock.request_history
        req_count = _count_mock_history(history, self.nfvo_url)
        self.assertEqual(1, req_count)

    def test_download_vnf_packages_with_retry_raise_not_found(self):

        fetch_base_url = os.path.join(self.url, uuidsentinel.vnf_pkg_id)
        self.requests_mock.register_uri(
            'GET',
            os.path.join(
                fetch_base_url,
                'package_content'),
            headers=self.headers,
            status_code=404)

        try:
            nfvo_client.VnfPackageRequest.download_vnf_packages(
                uuidsentinel.vnf_pkg_id)
        except requests.exceptions.RequestException as e:
            self.assertEqual(404, e.response.status_code)

        history = self.requests_mock.request_history
        req_count = _count_mock_history(history, self.nfvo_url)
        self.assertEqual(
            cfg.CONF.connect_vnf_packages.retry_num + 1, req_count)

    def test_download_vnf_packages_with_retry_raise_timeout(self):

        fetch_base_url = os.path.join(self.url, uuidsentinel.vnf_pkg_id)
        self.requests_mock.register_uri(
            'GET',
            os.path.join(
                fetch_base_url,
                'package_content'),
            exc=requests.exceptions.ConnectTimeout)

        try:
            nfvo_client.VnfPackageRequest.download_vnf_packages(
                uuidsentinel.vnf_pkg_id)
        except requests.exceptions.RequestException as e:
            self.assertIsNone(e.response)

        history = self.requests_mock.request_history
        req_count = _count_mock_history(history, self.nfvo_url)
        self.assertEqual(
            cfg.CONF.connect_vnf_packages.retry_num + 1, req_count)

    def test_download_vnf_packages_raise_failed_download_content(self):

        fetch_base_url = os.path.join(self.url, uuidsentinel.vnf_pkg_id)
        self.requests_mock.register_uri('GET', os.path.join(
            fetch_base_url, 'package_content'), content=None)

        self.assertRaises(
            nfvo_client.FaliedDownloadContentException,
            nfvo_client.VnfPackageRequest.download_vnf_packages,
            uuidsentinel.vnf_pkg_id)

        history = self.requests_mock.request_history
        req_count = _count_mock_history(history, self.nfvo_url)
        self.assertEqual(1, req_count)

    @ddt.data(None, "", " ")
    def test_download_vnf_packages_raise_non_baseurl(self, empty_val):
        cfg.CONF.set_override("base_url", empty_val,
                              group='connect_vnf_packages')

        self.assertRaises(
            nfvo_client.UndefinedExternalSettingException,
            nfvo_client.VnfPackageRequest.download_vnf_packages,
            uuidsentinel.vnf_pkg_id)

    @ddt.data(None, [], ["non"])
    def test_download_vnf_packages_raise_non_pipeline(self, empty_val):
        cfg.CONF.set_override('pipeline', empty_val,
                              group='connect_vnf_packages')

        self.assertRaises(
            nfvo_client.UndefinedExternalSettingException,
            nfvo_client.VnfPackageRequest.download_vnf_packages,
            uuidsentinel.vnf_pkg_id)

    def test_index(self):
        response_body = self.json_serial_date_to_dict(
            [fakes.VNFPACKAGE_RESPONSE, fakes.VNFPACKAGE_RESPONSE])
        self.requests_mock.register_uri(
            'GET', self.url, headers=self.headers, json=response_body)

        res = nfvo_client.VnfPackageRequest.index()
        self.assertEqual(200, res.status_code)
        self.assertIsInstance(res.json(), list)
        self.assertEqual(response_body, res.json())
        self.assertEqual(2, len(res.json()))
        self.assertEqual(response_body, json.loads(res.text))

        history = self.requests_mock.request_history
        req_count = _count_mock_history(history, self.nfvo_url)
        self.assertEqual(1, req_count)

    def test_index_with_auth_basic(self):
        cfg.CONF.set_override('auth_type', 'BASIC',
                              group='authentication')
        auth.auth_manager = auth._AuthManager()

        response_body = self.json_serial_date_to_dict(
            [fakes.VNFPACKAGE_RESPONSE, fakes.VNFPACKAGE_RESPONSE])
        self.requests_mock.register_uri(
            'GET', self.url, headers=self.headers, json=response_body)

        res = nfvo_client.VnfPackageRequest.index()
        self.assertEqual(200, res.status_code)
        self.assertIsInstance(res.json(), list)
        self.assertEqual(response_body, res.json())
        self.assertEqual(2, len(res.json()))
        self.assertEqual(response_body, json.loads(res.text))

        history = self.requests_mock.request_history
        req_count = _count_mock_history(history, self.nfvo_url)
        self.assertEqual(1, req_count)
        self.assert_auth_basic(history[0])

    def test_index_with_auth_client_credentials(self):
        cfg.CONF.set_override('auth_type', 'OAUTH2_CLIENT_CREDENTIALS',
                              group='authentication')

        self.requests_mock.register_uri('POST',
            self.token_endpoint,
            json={'access_token': 'test_token', 'token_type': 'bearer'},
            headers={'Content-Type': 'application/json'},
            status_code=200)

        auth.auth_manager = auth._AuthManager()

        response_body = self.json_serial_date_to_dict(
            [fakes.VNFPACKAGE_RESPONSE, fakes.VNFPACKAGE_RESPONSE])
        self.requests_mock.register_uri(
            'GET', self.url, headers=self.headers, json=response_body)

        res = nfvo_client.VnfPackageRequest.index()
        self.assertEqual(200, res.status_code)
        self.assertIsInstance(res.json(), list)
        self.assertEqual(response_body, res.json())
        self.assertEqual(2, len(res.json()))
        self.assertEqual(response_body, json.loads(res.text))

        history = self.requests_mock.request_history
        req_count = _count_mock_history(history, self.nfvo_url, self.oauth_url)
        self.assertEqual(2, req_count)
        self.assert_auth_basic(history[0])
        self.assert_auth_client_credentials(history[1], "test_token")

    def test_index_raise_not_found(self):
        self.requests_mock.register_uri(
            'GET', self.url, headers=self.headers, status_code=404)

        try:
            nfvo_client.VnfPackageRequest.index()
        except requests.exceptions.RequestException as e:
            self.assertEqual(404, e.response.status_code)

        history = self.requests_mock.request_history
        req_count = _count_mock_history(history, self.nfvo_url)
        self.assertEqual(
            cfg.CONF.connect_vnf_packages.retry_num + 1, req_count)

    def test_index_raise_non_baseurl(self):
        cfg.CONF.set_override("base_url", None,
                              group='connect_vnf_packages')

        self.assertRaises(nfvo_client.UndefinedExternalSettingException,
                          nfvo_client.VnfPackageRequest.index)

    def test_show(self):
        response_body = self.json_serial_date_to_dict(
            fakes.VNFPACKAGE_RESPONSE)
        self.requests_mock.register_uri(
            'GET',
            os.path.join(
                self.url,
                uuidsentinel.vnf_pkg_id),
            headers=self.headers,
            json=response_body)

        res = nfvo_client.VnfPackageRequest.show(uuidsentinel.vnf_pkg_id)
        self.assertEqual(200, res.status_code)
        self.assertIsInstance(res.json(), dict)
        self.assertEqual(response_body, res.json())
        self.assertEqual(response_body, json.loads(res.text))

        history = self.requests_mock.request_history
        req_count = _count_mock_history(history, self.nfvo_url)
        self.assertEqual(1, req_count)

    def test_show_with_auth_basic(self):
        cfg.CONF.set_override('auth_type', 'BASIC',
                              group='authentication')
        auth.auth_manager = auth._AuthManager()

        response_body = self.json_serial_date_to_dict(
            fakes.VNFPACKAGE_RESPONSE)
        self.requests_mock.register_uri(
            'GET',
            os.path.join(
                self.url,
                uuidsentinel.vnf_pkg_id),
            headers=self.headers,
            json=response_body)

        res = nfvo_client.VnfPackageRequest.show(uuidsentinel.vnf_pkg_id)
        self.assertEqual(200, res.status_code)
        self.assertIsInstance(res.json(), dict)
        self.assertEqual(response_body, res.json())
        self.assertEqual(response_body, json.loads(res.text))

        history = self.requests_mock.request_history
        req_count = _count_mock_history(history, self.nfvo_url)
        self.assertEqual(1, req_count)
        self.assert_auth_basic(history[0])

    def test_show_with_auth_client_credentials(self):
        cfg.CONF.set_override('auth_type', 'OAUTH2_CLIENT_CREDENTIALS',
                              group='authentication')

        self.requests_mock.register_uri('POST',
            self.token_endpoint,
            json={'access_token': 'test_token', 'token_type': 'bearer'},
            headers={'Content-Type': 'application/json'},
            status_code=200)

        auth.auth_manager = auth._AuthManager()

        response_body = self.json_serial_date_to_dict(
            fakes.VNFPACKAGE_RESPONSE)
        self.requests_mock.register_uri(
            'GET',
            os.path.join(
                self.url,
                uuidsentinel.vnf_pkg_id),
            headers=self.headers,
            json=response_body)

        res = nfvo_client.VnfPackageRequest.show(uuidsentinel.vnf_pkg_id)
        self.assertEqual(200, res.status_code)
        self.assertIsInstance(res.json(), dict)
        self.assertEqual(response_body, res.json())
        self.assertEqual(response_body, json.loads(res.text))

        history = self.requests_mock.request_history
        req_count = _count_mock_history(history, self.nfvo_url, self.oauth_url)
        self.assertEqual(2, req_count)
        self.assert_auth_basic(history[0])
        self.assert_auth_client_credentials(history[1], "test_token")

    def test_show_raise_not_found(self):
        self.requests_mock.register_uri(
            'GET',
            os.path.join(
                self.url,
                uuidsentinel.vnf_pkg_id),
            headers=self.headers,
            status_code=404)

        try:
            nfvo_client.VnfPackageRequest.show(uuidsentinel.vnf_pkg_id)
        except requests.exceptions.RequestException as e:
            self.assertEqual(404, e.response.status_code)

        history = self.requests_mock.request_history
        req_count = _count_mock_history(history, self.nfvo_url)
        self.assertEqual(
            cfg.CONF.connect_vnf_packages.retry_num + 1, req_count)

    def test_show_raise_non_baseurl(self):
        cfg.CONF.set_override("base_url", None,
                              group='connect_vnf_packages')

        self.assertRaises(nfvo_client.UndefinedExternalSettingException,
                          nfvo_client.VnfPackageRequest.show,
                          uuidsentinel.vnf_pkg_id)


@ddt.ddt
class TestGrantRequest(base.BaseTestCase):

    def setUp(self):
        super(TestGrantRequest, self).setUp()
        self.requests_mock = self.useFixture(requests_mock_fixture.Fixture())
        self.url = "http://nfvo.co.jp/grant/v1/grants"
        self.nfvo_url = 'http://nfvo.co.jp'
        self.headers = {'content-type': 'application/json'}

        self.token_endpoint = 'https://oauth2/tokens'
        self.nfvo_url = 'http://nfvo.co.jp'
        self.oauth_url = 'https://oauth2'
        self.auth_user_name = 'test_user'
        self.auth_password = 'test_password'

        cfg.CONF.set_override('auth_type', None,
                              group='authentication')
        cfg.CONF.set_override("base_url", self.url, group='connect_grant')
        cfg.CONF.set_override('user_name', self.auth_user_name,
                              group='authentication')
        cfg.CONF.set_override('password', self.auth_password,
                              group='authentication')
        cfg.CONF.set_override('token_endpoint', self.token_endpoint,
                              group='authentication')
        cfg.CONF.set_override('client_id', self.auth_user_name,
                              group='authentication')
        cfg.CONF.set_override('client_password', self.auth_password,
                              group='authentication')
        auth.auth_manager = auth._AuthManager()
        nfvo_client.GrantRequest._connector = nfvo_client._Connect(2, 1, 20)

    def tearDown(self):
        super(TestGrantRequest, self).tearDown()
        self.addCleanup(mock.patch.stopall)

    def assert_auth_basic(self, acutual_request):
        actual_auth = acutual_request._request.headers.get("Authorization")
        expected_auth = base64.b64encode(
            '{}:{}'.format(
                self.auth_user_name,
                self.auth_password).encode('utf-8')).decode()
        self.assertEqual("Basic " + expected_auth, actual_auth)

    def assert_auth_client_credentials(self, acutual_request, expected_token):
        actual_auth = acutual_request._request.headers.get(
            "Authorization")
        self.assertEqual("Bearer " + expected_token, actual_auth)

    def create_request_body(self):
        return {
            "vnfInstanceId": uuidsentinel.vnf_instance_id,
            "vnfLcmOpOccId": uuidsentinel.vnf_lcm_op_occ_id,
            "operation": "INST",
            "isAutomaticInvocation": False,
            "links": {
                "vnfLcmOpOcc": {
                    "href":
                        "https://localost/vnfm/vnflcm/v1/vnf_lcm_op_occs/" +
                    uuidsentinel.vnf_lcm_op_occ_id},
                "vnfInstance": {
                    "href": "https://localost/vnfm/vnflcm/v1/vnf_instances/" +
                    uuidsentinel.vnf_instance_id}}}

    def fake_response_body(self):
        return {
            "id": uuidsentinel.grant_id,
            "vnfInstanceId": uuidsentinel.vnf_instance_id,
            "vnfLcmOpOccId": uuidsentinel.vnf_lcm_op_occ_id,
            "additionalParams": {},
            "_links": {
                "self": {
                    "href":
                        "http://nfvo.co.jp/grant/v1/grants/\
                            19533fd4-eacb-4e6f-acd9-b56210a180d7"},
                "vnfLcmOpOcc": {
                    "href":
                        "https://localost/vnfm/vnflcm/v1/vnf_lcm_op_occs/" +
                    uuidsentinel.vnf_lcm_op_occ_id},
                "vnfInstance": {
                    "href": "https://localost/vnfm/vnflcm/v1/vnf_instances/" +
                    uuidsentinel.vnf_instance_id}}}

    def test_init(self):
        self.assertEqual(self.url, cfg.CONF.connect_grant.base_url)
        self.assertEqual(2, cfg.CONF.connect_grant.retry_num)
        self.assertEqual(30, cfg.CONF.connect_grant.retry_wait)
        self.assertEqual(20, cfg.CONF.connect_grant.timeout)

    def test_grants(self):
        response_body = self.fake_response_body()
        self.requests_mock.register_uri(
            'POST',
            self.url,
            json=response_body,
            headers=self.headers,
            status_code=201)

        request_body = self.create_request_body()
        res = nfvo_client.GrantRequest.grants(data=request_body)
        self.assertEqual(response_body, json.loads(res.text))
        self.assertEqual(response_body, res.json())

        history = self.requests_mock.request_history
        req_count = _count_mock_history(history, self.nfvo_url)
        self.assertEqual(1, req_count)

    def test_grants_with_retry_raise_bad_request(self):
        response_body = self.fake_response_body()
        self.requests_mock.register_uri('POST', self.url, json=json.dumps(
            response_body), headers=self.headers, status_code=400)

        request_body = self.create_request_body()
        try:
            nfvo_client.GrantRequest.grants(data=request_body)
        except requests.exceptions.RequestException as e:
            self.assertEqual(400, e.response.status_code)

        history = self.requests_mock.request_history
        req_count = _count_mock_history(history, self.nfvo_url)
        self.assertEqual(
            cfg.CONF.connect_grant.retry_num + 1, req_count)

    def test_grants_with_retry_raise_timeout(self):
        self.requests_mock.register_uri(
            'POST', self.url, exc=requests.exceptions.ConnectTimeout)

        request_body = self.create_request_body()
        try:
            nfvo_client.GrantRequest.grants(data=request_body)
        except requests.exceptions.RequestException as e:
            self.assertIsNone(e.response)

        history = self.requests_mock.request_history
        req_count = _count_mock_history(history, self.nfvo_url)
        self.assertEqual(
            cfg.CONF.connect_grant.retry_num + 1, req_count)

    @ddt.data(None, "", " ")
    def test_grants_raise_non_baseurl(self, empty_val):
        cfg.CONF.set_override("base_url", empty_val, group='connect_grant')
        self.assertRaises(nfvo_client.UndefinedExternalSettingException,
                          nfvo_client.GrantRequest.grants,
                          data={"test": "value1"})

    def test_grants_with_auth_basic(self):
        cfg.CONF.set_override('auth_type', 'BASIC',
                              group='authentication')
        auth.auth_manager = auth._AuthManager()

        response_body = self.fake_response_body()
        self.requests_mock.register_uri(
            'POST',
            self.url,
            json=response_body,
            headers=self.headers,
            status_code=201)

        request_body = self.create_request_body()
        res = nfvo_client.GrantRequest.grants(data=request_body)
        self.assertEqual(response_body, json.loads(res.text))
        self.assertEqual(response_body, res.json())

        history = self.requests_mock.request_history
        req_count = _count_mock_history(history, self.nfvo_url)
        self.assertEqual(1, req_count)
        self.assert_auth_basic(history[0])

    def test_grants_with_auth_client_credentials(self):
        cfg.CONF.set_override('auth_type', 'OAUTH2_CLIENT_CREDENTIALS',
                              group='authentication')

        self.requests_mock.register_uri('POST',
            self.token_endpoint,
            json={'access_token': 'test_token', 'token_type': 'bearer'},
            headers={'Content-Type': 'application/json'},
            status_code=200)

        auth.auth_manager = auth._AuthManager()

        response_body = self.fake_response_body()
        self.requests_mock.register_uri(
            'POST',
            self.url,
            json=response_body,
            headers=self.headers,
            status_code=201)

        request_body = self.create_request_body()
        res = nfvo_client.GrantRequest.grants(data=request_body)
        self.assertEqual(response_body, json.loads(res.text))
        self.assertEqual(response_body, res.json())

        history = self.requests_mock.request_history
        req_count = _count_mock_history(history, self.nfvo_url, self.oauth_url)
        self.assertEqual(2, req_count)
        self.assert_auth_basic(history[0])
        self.assert_auth_client_credentials(history[1], "test_token")



# Copyright 2016 Brocade Communications System, Inc.
# All Rights Reserved.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sqlalchemy as sa

from tacker.db import model_base
from tacker.db import types


class Event(model_base.BASE):
    id = sa.Column(sa.Integer, primary_key=True, nullable=False,
                   autoincrement=True)
    resource_id = sa.Column(types.Uuid, nullable=False)
    resource_state = sa.Column(sa.String(64), nullable=False)
    resource_type = sa.Column(sa.String(64), nullable=False)
    timestamp = sa.Column(sa.DateTime, nullable=False)
    event_type = sa.Column(sa.String(64), nullable=False)
    event_details = sa.Column(types.Json)



# Copyright 2010-2011 OpenStack Foundation
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Base Test Case for all Unit Tests"""

import contextlib
import gc
import logging
import os
import os.path
from unittest import mock
import weakref

import eventlet.timeout
import fixtures
from oslo_config import cfg
from oslo_messaging import conffixture as messaging_conffixture
import testtools

from tacker.common import config
from tacker.common import rpc as n_rpc
from tacker import context
from tacker import manager
from tacker import policy
from tacker.tests import fake_notifier
from tacker.tests import post_mortem_debug

CONF = cfg.CONF
CONF.import_opt('state_path', 'tacker.common.config')
TRUE_STRING = ['True', '1']
LOG_FORMAT = "%(asctime)s %(levelname)8s [%(name)s] %(message)s"

ROOTDIR = os.path.dirname(__file__)
ETCDIR = os.path.join(ROOTDIR, 'etc')


def etcdir(*p):
    return os.path.join(ETCDIR, *p)


def fake_use_fatal_exceptions(*args):
    return True


def fake_consume_in_threads(self):
    return []


class BaseTestCase(testtools.TestCase):

    def cleanup_core_plugin(self):
        """Ensure that the core plugin is deallocated."""
        nm = manager.TackerManager
        if not nm.has_instance():
            return

        # Perform a check for deallocation only if explicitly
        # configured to do so since calling gc.collect() after every
        # test increases test suite execution time by ~50%.
        check_plugin_deallocation = (
            os.environ.get('OS_CHECK_PLUGIN_DEALLOCATION') in TRUE_STRING)
        if check_plugin_deallocation:
            plugin = weakref.ref(nm._instance.plugin)

        nm.clear_instance()

        if check_plugin_deallocation:
            gc.collect()

            # TODO(marun) Ensure that mocks are deallocated?
            if plugin() and not isinstance(plugin(), mock.Base):
                self.fail('The plugin for this test was not deallocated.')

    def setup_coreplugin(self, core_plugin=None):
        if core_plugin is not None:
            cfg.CONF.set_override('core_plugin', core_plugin)

    def setup_notification_driver(self, notification_driver=None):
        self.addCleanup(fake_notifier.reset)
        if notification_driver is None:
            notification_driver = [fake_notifier.__name__]
        cfg.CONF.set_override("notification_driver", notification_driver)

    @staticmethod
    def config_parse(conf=None, args=None):
        """Create the default configurations."""
        # tacker.conf.test includes rpc_backend which needs to be cleaned up
        if args is None:
            args = ['--config-file', etcdir('tacker.conf.test')]
        if conf is None:
            config.init(args=args)
        else:
            conf(args)

    def setup_config(self, args=None):
        """Tests that need a non-default config can override this method."""
        self.config_parse(args=args)

    def setUp(self):
        super(BaseTestCase, self).setUp()

        # Ensure plugin cleanup is triggered last so that
        # test-specific cleanup has a chance to release references.
        self.addCleanup(self.cleanup_core_plugin)

        # Configure this first to ensure pm debugging support for setUp()
        if os.environ.get('OS_POST_MORTEM_DEBUG') in TRUE_STRING:
            self.addOnException(post_mortem_debug.exception_handler)

        if os.environ.get('OS_DEBUG') in TRUE_STRING:
            _level = logging.DEBUG
        else:
            _level = logging.INFO
        capture_logs = os.environ.get('OS_LOG_CAPTURE') in TRUE_STRING
        if not capture_logs:
            logging.basicConfig(format=LOG_FORMAT, level=_level)
        self.log_fixture = self.useFixture(
            fixtures.FakeLogger(
                format=LOG_FORMAT,
                level=_level,
                nuke_handlers=capture_logs,
            ))

        # suppress all but errors here
        self.useFixture(
            fixtures.FakeLogger(
                name='tacker.api.extensions',
                format=LOG_FORMAT,
                level=logging.ERROR,
                nuke_handlers=capture_logs,
            ))

        test_timeout = int(os.environ.get('OS_TEST_TIMEOUT', 0))
        if test_timeout == -1:
            test_timeout = 0
        if test_timeout > 0:
            self.useFixture(fixtures.Timeout(test_timeout, gentle=True))

        # If someone does use tempfile directly, ensure that it's cleaned up
        self.useFixture(fixtures.NestedTempfile())
        self.useFixture(fixtures.TempHomeDir())

        self.temp_dir = self.useFixture(fixtures.TempDir()).path
        cfg.CONF.set_override('state_path', self.temp_dir)

        self.setup_config()
        policy.init()
        self.addCleanup(policy.reset)
        self.addCleanup(mock.patch.stopall)
        self.addCleanup(CONF.reset)

        if os.environ.get('OS_STDOUT_CAPTURE') in TRUE_STRING:
            stdout = self.useFixture(fixtures.StringStream('stdout')).stream
            self.useFixture(fixtures.MonkeyPatch('sys.stdout', stdout))
        if os.environ.get('OS_STDERR_CAPTURE') in TRUE_STRING:
            stderr = self.useFixture(fixtures.StringStream('stderr')).stream
            self.useFixture(fixtures.MonkeyPatch('sys.stderr', stderr))
        self.useFixture(fixtures.MonkeyPatch(
            'tacker.common.exceptions.TackerException.use_fatal_exceptions',
            fake_use_fatal_exceptions))

        self.useFixture(fixtures.MonkeyPatch(
            'oslo_messaging.Notifier', fake_notifier.FakeNotifier))

        self.messaging_conf = messaging_conffixture.ConfFixture(CONF)
        self.messaging_conf.transport_url = 'fake:/'
        self.messaging_conf.response_timeout = 15
        self.useFixture(self.messaging_conf)

        self.addCleanup(n_rpc.clear_extra_exmods)
        n_rpc.add_extra_exmods('tacker.test')

        self.addCleanup(n_rpc.cleanup)
        n_rpc.init(CONF)

    def fake_admin_context(self):
        return context.get_admin_context()

    def config(self, **kw):
        """Override some configuration values.

        The keyword arguments are the names of configuration options to
        override and their values.

        If a group argument is supplied, the overrides are applied to
        the specified configuration option group.

        All overrides are automatically cleared at the end of the current
        test by the fixtures cleanup process.
        """
        group = kw.pop('group', None)
        for k, v in (kw).items():
            CONF.set_override(k, v, group)

    @contextlib.contextmanager
    def assert_max_execution_time(self, max_execution_time=5):
        with eventlet.timeout.Timeout(max_execution_time, False):
            yield
            return
        self.fail('Execution of this test timed out')

    def get_new_temp_dir(self):
        """Create a new temporary directory.

        :returns: fixtures.TempDir
        """
        return self.useFixture(fixtures.TempDir())

    def get_default_temp_dir(self):
        """Create a default temporary directory.

        Returns the same directory during the whole test case.

        :returns: fixtures.TempDir
        """
        if not hasattr(self, '_temp_dir'):
            self._temp_dir = self.get_new_temp_dir()
        return self._temp_dir

    def get_temp_file_path(self, filename, root=None):
        """Returns an absolute path for a temporary file.

        If root is None, the file is created in default temporary directory. It
        also creates the directory if it's not initialized yet.

        If root is not None, the file is created inside the directory passed as
        root= argument.

        :param filename: filename
        :type filename: string
        :param root: temporary directory to create a new file in
        :type root: fixtures.TempDir
        :returns: absolute file path string
        """
        root = root or self.get_default_temp_dir()
        return root.join(filename)



# Copyright 2016 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""enable soft delete

Revision ID: 941b5a6fff9e
Revises: 2ff0a0e360f1
Create Date: 2016-06-06 10:12:49.787430

"""

# flake8: noqa: E402

# revision identifiers, used by Alembic.
revision = '941b5a6fff9e'
down_revision = '2ff0a0e360f1'

from alembic import op
import sqlalchemy as sa


def upgrade(active_plugins=None, options=None):
    for table in ['vims', 'vnf', 'vnfd']:
        op.add_column(table,
                      sa.Column('deleted_at', sa.DateTime(), nullable=True))

    # unique constraint is taken care by the nfvo_db plugin to support
    # soft deletion of vim
    op.drop_index('auth_url', table_name='vimauths')



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from tacker.sol_refactored.api import wsgi as sol_wsgi
from tacker.sol_refactored.common import exceptions as sol_ex
from tacker.tests.unit import base


class TestWsgi(base.TestCase):

    def test_response_too_big(self):
        self.config_fixture.config(group='v2_vnfm', max_content_length=20)
        body = {"key": "value0123456789"}
        response = sol_wsgi.SolResponse(200, body)
        self.assertRaises(sol_ex.ResponseTooBig,
            response.serialize, mock.Mock(), 'application/json')

    def test_unknown_error_response(self):
        err_msg = "Test error"
        status = 500
        response = sol_wsgi.SolErrorResponse(Exception(err_msg), mock.Mock())
        problem_details = {
            "status": status,
            "detail": err_msg
        }
        self.assertEqual(status, response.status)
        self.assertEqual(problem_details, response.body)

    def test_check_api_version_no_version(self):
        resource = sol_wsgi.SolResource(mock.Mock())
        request = mock.Mock()
        request.headers = {}
        self.assertRaises(sol_ex.APIVersionMissing,
            resource.check_api_version, request)



# Copyright (c) 2022 FUJITSU
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from kubernetes import client
from samples.mgmt_driver.kubernetes.container_update import (
    container_update_mgmt as mgmt_driver)
from tacker.common import exceptions
from tacker import context
from tacker.tests.unit import base
from tacker.tests.unit.vnflcm import fakes
from tacker.tests.unit.vnfm.mgmt_drivers import fakes as mgmt_fakes
from tacker.vnflcm import utils as vnflcm_utils
from tacker.vnfm.infra_drivers.kubernetes.kubernetes_driver import Kubernetes
from unittest import mock


class FakeVimClient(mock.Mock):
    pass


class TestContainerUpdate(base.TestCase):

    def setUp(self):
        super(TestContainerUpdate, self).setUp()
        self.context = context.get_admin_context()
        self.cntr_update_mgmt = mgmt_driver.ContainerUpdateMgmtDriver()
        self.vnf_instance = mgmt_fakes.get_vnf_instance_object()
        self.modify_vnf_request = None
        self._mock_vim_client()
        self._stub_get_vim()
        self._mock_get_vnf_package_path()
        self.yaml_path_before = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "../../../etc/samples/etsi/nfv/"
            "test_cnf_container_update_before/")

    def _mock_vim_client(self):
        self.vim_client = mock.Mock(wraps=FakeVimClient())
        fake_vim_client = mock.Mock()
        fake_vim_client.return_value = self.vim_client
        self._mock(
            'tacker.vnfm.vim_client.VimClient', fake_vim_client)

    def _stub_get_vim(self):
        vim_obj = {'vim_id': '15053249-6979-8ee5-67d2-189fa9379534',
                   'vim_name': 'fake_vim', 'vim_auth':
                       {'auth_url': 'http://localhost/identity', 'password':
                           'test_pw', 'username': 'test_user', 'project_name':
                           'test_project'}, 'vim_type': 'openstack', 'tenant':
                       '15053249-6979-8ee5-67d2-189fa9379534'}
        self.vim_client.get_vim.return_value = vim_obj

    def _mock_get_vnf_package_path(self):
        vnflcm_utils.get_vnf_package_path = mock.Mock(
            return_value=os.path.join(
                os.path.abspath(os.path.dirname(__file__)),
                "../../../etc/samples/etsi/nfv/"
                "test_cnf_container_update_after/")
        )

    def test_container_update_get_type(self):
        get_type = self.cntr_update_mgmt.get_type()
        self.assertEqual('mgmt-container-update', get_type)

    def test_container_update_get_name(self):
        get_name = self.cntr_update_mgmt.get_name()
        self.assertEqual('mgmt-container-update', get_name)

    def test_container_update_get_description(self):
        get_description = self.cntr_update_mgmt.get_description()
        self.assertEqual(
            'Tacker Container Update VNF Mgmt Driver', get_description)

    def test_container_update_modify_information_start(self):
        modify_start = self.cntr_update_mgmt.modify_information_start(
            self.context, self.vnf_instance, self.modify_vnf_request)
        self.assertIsNone(modify_start)

    def test_container_update_modify_container_img(self):
        old_containers = [
            client.V1Container(image="curry", name="curry")
        ]

        new_containers = [
            client.V1Container(image="curry1", name="curry")
        ]
        self.cntr_update_mgmt._modify_container_img(
            old_containers, new_containers)
        self.assertEqual('curry1', old_containers[0].image)

    @mock.patch.object(client.CoreV1Api, 'list_namespaced_pod')
    def test_container_update_replace_wait_k8s_timeout(self, mock_list_pod):
        k8s_pod_objs = [
            {'namespace': 'default', 'object': mgmt_fakes.fake_pod()}
        ]
        kube_driver = Kubernetes()
        kube_driver.STACK_RETRIES = 1
        kube_driver.STACK_RETRY_WAIT = 5
        mock_list_pod.return_value = mgmt_fakes.fake_list_pod()
        self.assertRaises(
            exceptions.MgmtDriverOtherError,
            self.cntr_update_mgmt._replace_wait_k8s, kube_driver,
            k8s_pod_objs, client.CoreV1Api, self.vnf_instance)

    @mock.patch.object(client.CoreV1Api, 'list_namespaced_pod')
    def test_container_update_replace_wait_k8s_unknown(self, mock_list_pod):
        k8s_pod_objs = [
            {'namespace': 'default', 'object': mgmt_fakes.fake_pod()}
        ]
        kube_driver = Kubernetes()
        pod_list = mgmt_fakes.fake_list_pod()
        pod_list.items[0].status.phase = 'Unknown'
        mock_list_pod.return_value = pod_list
        self.assertRaises(
            exceptions.MgmtDriverOtherError,
            self.cntr_update_mgmt._replace_wait_k8s, kube_driver,
            k8s_pod_objs, client.CoreV1Api, self.vnf_instance)

    @mock.patch('tacker.objects.vnf_instance.VnfInstance.save')
    @mock.patch('tacker.vnflcm.utils._get_vnfd_dict')
    @mock.patch.object(client.CoreV1Api, 'list_namespaced_pod')
    @mock.patch.object(client.CoreV1Api, 'replace_namespaced_secret')
    @mock.patch.object(client.CoreV1Api, 'replace_namespaced_config_map')
    @mock.patch.object(client.CoreV1Api, 'replace_namespaced_pod')
    @mock.patch.object(client.CoreV1Api, 'read_namespaced_pod')
    def test_container_update_modify_information_end(
            self, mock_read_pod, mock_replace_pod, mock_replace_config_map,
            mock_replace_secret, mock_list_pod, mock_vnfd_dict, mock_save):
        mock_read_pod.return_value = mgmt_fakes.fake_pod()
        mock_list_pod.return_value = client.V1PodList(items=[
            mgmt_fakes.get_fake_pod_info(kind='Pod', name='vdu1')])
        mock_vnfd_dict.return_value = fakes.vnfd_dict_cnf()
        kwargs = {
            'old_vnf_package_path': self.yaml_path_before,
            'configmap_secret_paths': [
                "Files/kubernetes/configmap_2.yaml",
                "Files/kubernetes/secret_2.yaml"
            ]
        }
        self.cntr_update_mgmt.modify_information_end(
            self.context, self.vnf_instance, self.modify_vnf_request, **kwargs)
        self.assertEqual(1, mock_save.call_count)



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tacker.sol_refactored.objects.common import fields  # noqa

# NOTE: You may scratch your head as you see code that imports
# this module and then accesses attributes for objects such as Instance,
# etc, yet you do not see these attributes in here. Never fear, there is
# a little bit of magic. When objects are registered, an attribute is set
# on this module automatically, pointing to the newest/latest version of
# the object.


def register_all():
    # NOTE: You must make sure your object gets imported in this
    # function in order for it to be registered by services that may
    # need to receive it via RPC.
    objects_root = 'tacker.sol_refactored.objects'
    __import__(objects_root + '.common.cp_protocol_data')
    __import__(objects_root + '.common.ext_link_port_data')
    __import__(objects_root + '.common.ext_managed_virtual_link_data')
    __import__(objects_root + '.common.ext_virtual_link_data')
    __import__(objects_root + '.common.ip_over_ethernet_address_data')
    __import__(objects_root + '.common.link')
    __import__(objects_root + '.common.notification_link')
    __import__(objects_root + '.common.problem_details')
    __import__(objects_root + '.common.resource_handle')
    __import__(objects_root + '.common.subscription_authentication')
    __import__(objects_root + '.common.vim_connection_info')
    __import__(objects_root + '.common.vnf_ext_cp_config')
    __import__(objects_root + '.common.vnf_ext_cp_data')
    __import__(objects_root + '.common.vnf_instance_subscription_filter')
    __import__(objects_root + '.v1.constraint_resource_ref')
    __import__(objects_root + '.v1.fields')
    __import__(objects_root + '.v1.grant_info')
    __import__(objects_root + '.v1.grant')
    __import__(objects_root + '.v1.grant_request')
    __import__(objects_root + '.v1.placement_constraint')
    __import__(objects_root + '.v1.resource_definition')
    __import__(objects_root + '.v1.snapshot_resource_definition')
    __import__(objects_root + '.v1.vim_compute_resource_flavour')
    __import__(objects_root + '.v1.vim_constraint')
    __import__(objects_root + '.v1.vim_snapshot_resource')
    __import__(objects_root + '.v1.vim_software_image')
    __import__(objects_root + '.v1.zone_group_info')
    __import__(objects_root + '.v1.zone_info')
    __import__(objects_root + '.v2.affected_ext_link_port')
    __import__(objects_root + '.v2.affected_virtual_link')
    __import__(objects_root + '.v2.affected_virtual_storage')
    __import__(objects_root + '.v2.affected_vnfc')
    __import__(objects_root + '.v2.cancel_mode')
    __import__(objects_root + '.v2.change_current_vnf_pkg_request')
    __import__(objects_root + '.v2.change_ext_vnf_connectivity_request')
    __import__(objects_root + '.v2.change_vnf_flavour_request')
    __import__(objects_root + '.v2.cp_protocol_info')
    __import__(objects_root + '.v2.create_vnf_pkg_info_request')
    __import__(objects_root + '.v2.create_vnf_request')
    __import__(objects_root + '.v2.create_vnf_snapshot_info_request')
    __import__(objects_root + '.v2.create_vnf_snapshot_request')
    __import__(objects_root + '.v2.external_artifacts_access_config')
    __import__(objects_root + '.v2.ext_link_port_info')
    __import__(objects_root + '.v2.ext_managed_virtual_link_info')
    __import__(objects_root + '.v2.ext_virtual_link_info')
    __import__(objects_root + '.v2.fields')
    __import__(objects_root + '.v2.heal_vnf_request')
    __import__(objects_root + '.v2.instantiate_vnf_request')
    __import__(objects_root + '.v2.ip_over_ethernet_address_info')
    __import__(objects_root + '.v2.lccn_links')
    __import__(objects_root + '.v2.lccn_subscription')
    __import__(objects_root + '.v2.lccn_subscription_request')
    __import__(objects_root + '.v2.lifecycle_change_notifications_filter')
    __import__(objects_root + '.v2.modifications_triggered_by_vnf_pkg_change')
    __import__(objects_root + '.v2.monitoring_parameter')
    __import__(objects_root + '.v2.operate_vnf_request')
    __import__(objects_root + '.v2.pkgm_links')
    __import__(objects_root + '.v2.pkgm_notification_filter')
    __import__(objects_root + '.v2.pkgm_subscription_request')
    __import__(objects_root + '.v2.revert_to_vnf_snapshot_request')
    __import__(objects_root + '.v2.scale_info')
    __import__(objects_root + '.v2.scale_vnf_request')
    __import__(objects_root + '.v2.scale_vnf_to_level_request')
    __import__(objects_root + '.v2.terminate_vnf_request')
    __import__(objects_root + '.v2.upload_vnf_package_from_uri_request')
    __import__(objects_root + '.v2.virtual_storage_resource_info')
    __import__(objects_root + '.v2.vnfc_info')
    __import__(objects_root + '.v2.vnfc_info_modifications')
    __import__(objects_root + '.v2.vnfc_resource_info')
    __import__(objects_root + '.v2.vnfc_snapshot_info')
    __import__(objects_root + '.v2.vnf_ext_cp_info')
    __import__(objects_root + '.v2.vnf_identifier_creation_notification')
    __import__(objects_root + '.v2.vnf_identifier_deletion_notification')
    __import__(objects_root + '.v2.vnf_info_modification_request')
    __import__(objects_root + '.v2.vnf_info_modifications')
    __import__(objects_root + '.v2.vnf_instance')
    __import__(objects_root + '.v2.vnf_lcm_operation_occurrence_notification')
    __import__(objects_root + '.v2.vnf_lcm_op_occ')
    __import__(objects_root + '.v2.vnf_link_port_data')
    __import__(objects_root + '.v2.vnf_link_port_info')
    __import__(objects_root + '.v2.vnf_package_artifact_info')
    __import__(objects_root + '.v2.vnf_package_change_notification')
    __import__(objects_root + '.v2.vnf_package_onboarding_notification')
    __import__(objects_root + '.v2.vnf_package_software_image_info')
    __import__(objects_root + '.v2.vnf_pkg_info_modifications')
    __import__(objects_root + '.v2.vnf_pkg_info')
    __import__(objects_root + '.v2.vnf_snapshot_info_modification_request')
    __import__(objects_root + '.v2.vnf_snapshot_info_modifications')
    __import__(objects_root + '.v2.vnf_snapshot_info')
    __import__(objects_root + '.v2.vnf_snapshot')
    __import__(objects_root + '.v2.vnf_state_snapshot_info')
    __import__(objects_root + '.v2.vnf_virtual_link_resource_info')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log as logging
import paramiko

from tacker.common import exceptions

LOG = logging.getLogger(__name__)


class CommandResult(object):
    """Result class contains command, stdout, stderror and return code."""
    def __init__(self, cmd, stdout, stderr, return_code):
        self.__cmd = cmd
        self.__stdout = stdout
        self.__stderr = stderr
        self.__return_code = return_code

    def get_command(self):
        return self.__cmd

    def get_stdout(self):
        return self.__stdout

    def get_stderr(self):
        return self.__stderr

    def get_return_code(self):
        return self.__return_code

    def __str__(self):
        return "cmd: %s, stdout: %s, stderr: %s, return code: %s" \
            % (self.__cmd, self.__stdout, self.__stderr, self.__return_code)

    def __repr__(self):
        return "cmd: %s, stdout: %s, stderr: %s, return code: %s" \
            % (self.__cmd, self.__stdout, self.__stderr, self.__return_code)


class RemoteCommandExecutor(object):
    """Class to execute a command on remote location"""
    def __init__(self, user, password, host, timeout=10):
        self.__user = user
        self.__password = password
        self.__host = host
        self.__paramiko_conn = None
        self.__ssh = None
        self.__timeout = timeout
        self.__connect()

    def __connect(self):
        try:
            self.__ssh = paramiko.SSHClient()
            self.__ssh.set_missing_host_key_policy(paramiko.WarningPolicy())
            self.__ssh.connect(self.__host, username=self.__user,
                password=self.__password, timeout=self.__timeout)
            LOG.info("Connected to %s", self.__host)
        except paramiko.AuthenticationException:
            LOG.error("Authentication failed when connecting to %s",
                      self.__host)
            raise exceptions.NotAuthorized
        except paramiko.SSHException:
            LOG.error("Could not connect to %s. Giving up", self.__host)
            raise

    def close_session(self):
        self.__ssh.close()
        LOG.debug("Connection close")

    def execute_command(self, cmd, input_data=None):
        try:
            stdin, stdout, stderr = self.__ssh.exec_command(cmd)
            if input_data:
                stdin.write(input_data)
                LOG.debug("Input data written successfully")
                stdin.flush()
                LOG.debug("Input data flushed")
                stdin.channel.shutdown_write()

            # NOTE (dkushwaha): There might be a case, when server can take
            # too long time to write data in stdout buffer or sometimes hang
            # itself, in that case readlines() will stuck for long/infinite
            # time. To handle such cases, timeout logic should be introduce
            # here.
            cmd_out = stdout.readlines()
            cmd_err = stderr.readlines()
            return_code = stdout.channel.recv_exit_status()
        except paramiko.SSHException:
            LOG.error("Command execution failed at %s. Giving up", self.__host)
            raise
        result = CommandResult(cmd, cmd_out, cmd_err, return_code)
        LOG.debug("Remote command execution result: %s", result)
        return result

    def __del__(self):
        self.close_session()



# Copyright 2016 Brocade Communications System, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time
import yaml

from tacker.plugins.common import constants as evt_constants
from tacker.tests.functional import base
from tacker.tests.utils import read_file

from tackerclient.common import exceptions

SECRET_PASSWORD = '***'


class VimTestCreate(base.BaseTackerTest):

    def _test_create_delete_vim(self, vim_file, name, description, vim_type,
                                version=None):

        data, vim_arg = self._generate_vim_data(
            'local-vim.yaml', name, description, vim_type, version)
        # updated args
        new_name = "fake %s" % name
        new_desc = "fake %s" % description
        update_vim_arg = {'vim': {'name': new_name,
                                  'description': new_desc}}

        # Register vim
        vim_res = self.client.create_vim(vim_arg)
        vim_obj = vim_res['vim']
        vim_id = vim_obj['id']
        self.verify_vim(vim_obj, data, name, description, version)
        self.verify_vim_events(vim_id, evt_constants.RES_EVT_CREATE)

        # Read vim
        vim_show_res = self.client.show_vim(vim_id)
        self.verify_vim(vim_show_res['vim'], data, name, description, version)

        # Update vim
        vim_update = self.client.update_vim(vim_id, update_vim_arg)
        vim_obj = vim_update['vim']
        self.verify_vim(vim_obj, data, new_name, new_desc, version)
        self.verify_vim_events(vim_id, evt_constants.RES_EVT_UPDATE)

        # With the updated name above, create another VIM with the
        # same name and check for Duplicate name exception.
        vim_arg['vim']['name'] = update_vim_arg['vim']['name']
        msg = "vim already exist with given ['tenant_id', 'name', "\
              "'deleted_at']"
        try:
            self.client.create_vim(vim_arg)
        except Exception as err:
            self.assertEqual(err.message, msg)

        # Since there already exists a DEFAULT VM, Verify that a update
        # to is_default to TRUE for another VIM raises an exception.
        update_arg = {'vim': {'is_default': True}}
        msg = "Default VIM already exists."
        self.assertRaisesRegex(exceptions.InternalServerError, msg,
                               self.client.update_vim,
                               vim_id, update_arg)

        # Delete vim
        try:
            self.client.delete_vim(vim_id)
        except Exception:
            self.assertFalse(True, "Failed to delete vim %s" % vim_id)
        self.verify_vim_events(vim_id, evt_constants.RES_EVT_DELETE)

    def verify_vim(self, vim_instance, config_data, name, description,
                   version):
        expected_regions = ['RegionOne']
        self.assertIsNotNone(vim_instance)
        self.assertEqual(description, vim_instance['description'])
        self.assertEqual(name, vim_instance['name'])
        self.assertIsNotNone(vim_instance['tenant_id'])
        self.assertIsNotNone(vim_instance['id'])
        self.assertEqual(config_data['username'],
                         vim_instance['auth_cred']['username'])
        self.assertEqual(SECRET_PASSWORD,
                         vim_instance['auth_cred']['password'])
        self.assertEqual(expected_regions,
                         vim_instance['placement_attr']['regions'])
        if version:
            method_name = 'verify_vim_' + version
            getattr(self, method_name)(vim_instance, config_data)

    def verify_vim_events(self, vim_id, evt_type, tstamp=None, cnt=1):
        params = {'resource_id': vim_id,
                  'resource_type': evt_constants.RES_TYPE_VIM,
                  'event_type': evt_type}
        if tstamp:
            params['timestamp'] = tstamp

        vim_evt_list = self.client.list_vim_events(**params)

        self.assertIsNotNone(vim_evt_list['vim_events'],
                             "List of VIM events are Empty")
        self.assertEqual(cnt, len(vim_evt_list['vim_events']))

    def verify_vim_v3(self, vim_instance, config_data):
        self.assertEqual(config_data['project_name'],
                         vim_instance['auth_cred']['project_name'])

    def test_create_delete_local_vim(self):
        name = 'Default vim'
        description = 'Local vim description'
        vim_type = 'openstack'
        ks_version = 'v3'
        self._test_create_delete_vim('local-vim.yaml', name, description,
                                     vim_type, ks_version)

    def _generate_vim_data(self, vim_file, name, description, vim_type,
                           version=None):

        data = yaml.safe_load(read_file(vim_file))
        password = data['password']
        username = data['username']
        project_name = data['project_name']
        auth_url = data['auth_url']
        if version:
            if (not auth_url.endswith("/v3") or
               not auth_url.endswith("/v3/")):
                auth_url += "/v3"
        domain_name = data.get('domain_name', None)
        vim_arg = {'vim': {'name': name, 'description': description,
                           'type': vim_type,
                           'auth_url': auth_url,
                           'auth_cred': {'username': username,
                                         'password': password,
                                         'user_domain_name': domain_name},
                           'vim_project': {'name': project_name,
                                           'project_domain_name':
                                               domain_name},
                           'is_default': False}}
        return data, vim_arg

    def test_re_create_delete_local_vim(self):
        name = 'test_vim'
        description = 'Test vim description'
        vim_type = 'openstack'
        ks_version = 'v3'
        self._test_create_delete_vim('local-vim.yaml', name, description,
                                     vim_type, ks_version)
        time.sleep(1)
        self._test_create_delete_vim('local-vim.yaml', name, description,
                                     vim_type, ks_version)

    def test_duplicate_vim(self):
        name = 'test_duplicate_vim'
        description = 'Test duplicate vim description'
        vim_type = 'openstack'
        version = 'v3'
        data, vim_arg = self._generate_vim_data(
            'local-vim.yaml', name, description, vim_type, version)

        # Register vim
        vim_res = self.client.create_vim(vim_arg)
        vim_obj = vim_res['vim']
        vim_id = vim_obj['id']

        # Read vim
        vim_show_res = self.client.show_vim(vim_id)
        self.verify_vim(vim_show_res['vim'], data, name, description, version)
        msg = "vim already exist with given ['tenant_id', 'name', "\
              "'deleted_at']"
        err_msg = None
        try:
            self.client.create_vim(vim_arg)
        except Exception as err:
            err_msg = err.message
        self.assertEqual(err_msg, msg)
